"""
Various functions
-----------------
"""

from __future__ import annotations

from collections.abc import Iterator, MutableMapping
from datetime import datetime
from os import linesep
from pathlib import Path
from typing import TYPE_CHECKING, Any

import numpy as np

from pigreads._version import version as __version__

if TYPE_CHECKING:
    from pigreads.models import Models


def get_upper_triangle(
    matrix: np.ndarray[Any, Any],
) -> np.ndarray[Any, Any]:
    """
    Convert a 3x3 matrix to a 6D vector, with the diagonal and upper triangle
    of the matrix as elements in the order xx, yy, zz, yz, xz, xy. Additional
    dimensions are supported, but the last two dimensions must each have size
    3.

    :param matrix: A 3x3 matrix.
    :return: A 6D vector.
    """

    matrix = np.array(matrix)
    assert matrix.ndim >= 2, "matrix must have at least two dimensions"
    assert matrix.shape[-1] == 3, "matrix must be a 3x3 matrix (in the last two axes)"
    assert matrix.shape[-2] == 3, "matrix must be a 3x3 matrix (in the last two axes)"
    assert np.allclose(matrix[..., 1, 2], matrix[..., 2, 1]), (
        "The yz and zy components of matrix need to be the same"
    )
    assert np.allclose(matrix[..., 0, 2], matrix[..., 2, 0]), (
        "The xz and zx components of matrix need to be the same"
    )
    assert np.allclose(matrix[..., 0, 1], matrix[..., 1, 0]), (
        "The xy and yx components of matrix need to be the same"
    )
    triag: np.ndarray[Any, Any] = np.stack(
        (
            matrix[..., 0, 0],  # xx
            matrix[..., 1, 1],  # yy
            matrix[..., 2, 2],  # zz
            matrix[..., 1, 2],  # yz
            matrix[..., 0, 2],  # xz
            matrix[..., 0, 1],  # xy
        ),
        axis=-1,
    )
    return triag


def normalise_vector(
    f: np.ndarray[Any, Any] | list[int] | tuple[int, int, int],
    dtype: type = np.single,
) -> np.ndarray[Any, Any]:
    """
    Normalise a 3D vector to unit length.

    :param f: A 3D vector over space with shape (Nz, Ny, Nx, 3).
    :param dtype: Data type of the arrays, i.e., single or double precision floating point numbers.
    :return: A 5D vector with shape (Nz, Ny, Nx, 3, 1).
    """

    f = np.array(f, dtype=dtype)
    assert isinstance(f, np.ndarray)
    assert f.ndim >= 1, "f must be a 3D vector"
    assert f.ndim <= 4, "too many dimensions for f"
    assert f.shape[-1] == 3, "f must be a 3D vector (in the last axis)"
    while f.ndim < 4:
        f.shape = (1, *f.shape)
    norm = np.linalg.norm(f, axis=-1)
    nonzero = norm > 0
    norm.shape = (*norm.shape, 1)
    f[nonzero] /= norm[nonzero]
    f.shape = (*f.shape, 1)
    assert f.ndim == 5, "f must have 5 dimensions: z, y, x, row, col"
    assert f.shape[-1] == 1, "f must be a 3D column vector"
    assert f.shape[-2] == 3, "f must be a 3D column vector"
    return f


def to_ithildin(
    framedur: float,
    dt: float,
    dz: float,
    dy: float,
    dx: float,
    models: Models,
    states: np.ndarray[Any, Any],
    inhom: np.ndarray[Any, Any] | None = None,
) -> tuple[dict[str, Any], dict[str, np.ndarray[Any, Any]]]:
    """
    Convert the output of a Pigreads simulation to an Ithildin SimData object.

    While originally designed for a different reaction-diffusion solver,
    the Python module for Ithildin is useful to analyse Pigreads simulations.

    :param framedur: The duration between subsequent frames, usually in milliseconds.
    :param dt: The time step size used in :py:func:`pigreads.models.Models.run`,\
            usually in milliseconds.
    :param dz: The distance between points in the z-dimension,\
            see :py:func:`pigreads.helper.deltas`.
    :param dy: The distance between points in the y-dimension.
    :param dx: The distance between points in the x-dimension.
    :param models: The models used in the simulation,\
            see :py:class:`pigreads.models.Models`.
    :param states: The states of the simulation, a 5D array of shape \
            (Nt, Nz, Ny, Nx, Nv),\
            see :py:func:`pigreads.models.Models.resting_states`\
            and :py:func:`pigreads.models.Models.run`.
    :param inhom: A 3D array with integer values, encoding which model to use at each point. \
            Its value is zero for points outside the medium and one or more for points inside. \
            If ``None``, all points are considered inside the medium.
    :return: Tuple of an Ithildin log file as a string \
            and a dictionary of the variables by variable name.

    Usage::

        import ithildin as ith
        log, variables = pig.to_ithildin(Nt * dt, dt, dz, dy, dx, models, states, inhom)
        sd = ith.SimData(log=ith.Log(data=log))
        sd.vars = variables
    """

    Nt, Nz, Ny, Nx, _ = states.shape  # pylint: disable=invalid-name

    timestamp = datetime.now()

    log = {
        "Ithildin log version": 2,
        "Simulation parameters": {
            "Ithildin library version": f"pigreads {__version__}",
            "Timestep dt": dt,
            "Frame duration": framedur,
            "Number of frames to take": Nt,
            "Serial number": int(timestamp.strftime(r"%Y%m%d%H%M%S")),
            "Name of simulation series": "pigreads",
        },
        "Geometry parameters": {
            "Number of dimensions": 3,
            "Voxel size": [dx, dy, dz],
            "Domain size": [Nx, Ny, Nz],
        },
        "Start date": timestamp.isoformat(),
    }

    for i, model in enumerate(models):
        model_def = models.available[model.key]
        key = "Model parameters"
        if i > 0:
            key += f" {i}"
        log[key] = {
            "Model type": model_def.name,
            "Class": model.key,
            "Citation": linesep.join(model_def.dois),
            "Parameters": model_def(**model.parameters),
            "Initial values": model_def.variables,
            "Variable names": list(model_def.variables.keys()),
            "Number of vars": len(model_def.variables),
        }

    shape = (-1, Nz, Ny, Nx)
    variables: dict[str, np.ndarray[Any, Any]] = {
        v: states[..., iv].reshape(shape)
        for iv, v in enumerate(models.available[models[0].key].variables.keys())
    }
    if inhom is not None:
        variables["inhom"] = inhom.reshape(shape)

    return log, variables


def delta(x: np.ndarray[Any, Any], ax: int = -1) -> float:
    """
    Extract grid spacing from a 3D array.

    :param x: A 3D array.
    :param ax: The axis along which to calculate the distance.
    :return: The distance between the first two points.

    For example, consider this code::

        z, y, x = np.mgrid[0, 0:4:0.2, 0:1:5j]
        dx = pig.delta(z, ax=-1)
        dy = pig.delta(z, ax=-2)
        dz = pig.delta(z, ax=-3)
    """
    assert x.ndim == 3
    diff = np.diff(np.moveaxis(x, ax, -1)[0, 0, :2])
    return 1.0 if diff.shape[0] == 0 else float(diff[0])


def deltas(*x: np.ndarray[Any, Any]) -> list[float]:
    """
    Extract grid spacing from a 3D meshgrid.

    For example, consider this code::

        z, y, x = np.mgrid[0, 0:4:0.2, 0:1:5j]
        dz, dy, dx = pig.deltas(z, y, x)

    :param x: A 3D array.
    :return: A list with the distances between the points.
    """
    return [delta(xi, i) for i, xi in enumerate(x)]


def prepare_array(
    shape: tuple[int, ...],
    path: Path | str | None = None,
    dtype: type = np.single,
) -> np.ndarray[Any, Any]:
    """
    Prepare an array in a given shape.

    Either create a new array or load an existing array from the file with
    the given path as a memory map.

    The shape and dtype of the array are given as arguments. If the path is
    ``None``, a new array is created. If the path is a file, the array is
    loaded from the file. If the array is not of the correct shape or dtype
    or the file does not exist, a new array is created.

    The array is returned as a memory map if a path is given, otherwise as a
    normal numpy array.

    :param shape: Shape of the array.
    :param path: Path to the file to load the array from.
    :param dtype: Data type of the arrays, i.e., single or double precision floating point numbers.
    :return: Resulting array.
    :see: :py:func:`numpy.lib.format.open_memmap`
    """

    if path is None:
        return np.zeros(shape=shape, dtype=dtype)

    path = Path(path)
    if path.is_file():
        arr = np.lib.format.open_memmap(path, "r+")
        if arr.shape == shape and arr.dtype == dtype:
            return arr
        del arr

    arr = np.lib.format.open_memmap(
        path,
        "w+",
        dtype=dtype,
        shape=shape,
    )
    assert isinstance(arr, np.ndarray)
    arr[:] = np.nan
    return arr


class ModelParameters(MutableMapping[str, float]):
    """
    A view into the core implementation of the models
    allowing reading and modifying the parameters of a model.

    :param models: Instance of the models class to link to the core.
    :var models: Instance of the models class to link to the core.
    :vartype models: pigreads.models.Models
    :param imodel: Index of the model.
    :var imodel: Index of the model.
    :vartype imodel: int
    """

    def __init__(self, models: Models, imodel: int) -> None:
        self._models: Models = models
        self._imodel: int = imodel
        self._keys: list[str] = list(
            models.get_definition(imodel).all_parameters.keys()
        )

    def _key_to_index(self, key: str) -> int:
        """
        Get the index of the parameter with the given key.

        :param key: Key of the parameter.
        :return: Index of the parameter.

        .. include in docs
        """
        return next(i for i, k in enumerate(self._keys) if k == key)

    def __getitem__(self, key: str) -> float:
        """
        Get the parameter with the given key.

        :param key: Key of the parameter.
        :return: Parameter value.

        .. include in docs
        """
        return self._models.get_parameter(self._imodel, self._key_to_index(key))

    def __setitem__(self, key: str, value: float) -> None:
        """
        Set the parameter with the given key.

        :param key: Key of the parameter.
        :param value: New value of the parameter.

        .. include in docs
        """
        self._models.set_parameter(self._imodel, self._key_to_index(key), value)

    def __delitem__(self, key: str) -> None:
        """
        Delete the parameter with the given key.

        Note: This operation is not supported in this class.

        :param key: Key of the parameter.

        .. include in docs
        """
        message = "Deleting items is not supported in this class."
        raise NotImplementedError(message)

    def __iter__(self) -> Iterator[str]:
        """
        Get an iterator of the keys of the parameters.

        :return: Iterator of the keys of the parameters.

        .. include in docs
        """
        return iter(self._keys)

    def __len__(self) -> int:
        """
        Get the number of parameters.

        :return: Number of parameters.

        .. include in docs
        """
        return len(self._keys)

    def to_dict(self) -> dict[str, float]:
        """
        Convert the parameters to a dictionary.

        :return: Dictionary of the parameters.
        """
        return {k: self[k] for k in self._keys}

    def __repr__(self) -> str:
        """
        Get the string representation of the parameters.

        :return: String representation of the parameters.

        .. include in docs
        """
        return repr(self.to_dict())

    def __str__(self) -> str:
        """
        Get the string representation of the parameters.

        :return: String representation of the parameters.

        .. include in docs
        """
        return str(self.to_dict())


__all__ = [
    "ModelParameters",
    "delta",
    "deltas",
    "get_upper_triangle",
    "normalise_vector",
    "prepare_array",
    "to_ithildin",
]



from __future__ import annotations

import datetime
import importlib.metadata
from typing import Any

project = "Pigreads"
copyright = f"2024\u2013{datetime.datetime.now().year} Desmond Kabus"
author = "Desmond Kabus"
version = release = importlib.metadata.version("pigreads")

extensions = [
    "myst_parser",
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",
    "sphinx_copybutton",
    "sphinx_click",
    "sphinx_lfs_content",
    "sphinxcontrib.autodoc_pydantic",
]

source_suffix = [".rst", ".md"]
exclude_patterns = [
    "_build",
    "**.ipynb_checkpoints",
    "Thumbs.db",
    ".DS_Store",
    ".env",
    ".venv",
]

html_logo = "../logo.png"
html_favicon = "favicon.ico"
html_theme = "furo"

html_theme_options: dict[str, Any] = {
    "source_repository": "https://gitlab.com/pigreads/pigreads",
    "source_branch": "main",
    "source_directory": "docs/",
}

myst_enable_extensions = [
    "colon_fence",
]

intersphinx_mapping = {
    "click": ("https://click.palletsprojects.com/en/stable", None),
    "matplotlib": ("https://matplotlib.org/stable/", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "pydantic": ("https://docs.pydantic.dev/latest", None),
    "python": ("https://docs.python.org/3", None),
    "pyopencl": ("https://documen.tician.de/pyopencl", None),  # codespell: disable
}

nitpick_ignore = [
    ("py:class", "numpy.int32"),  # TODO: remove when possible
    ("py:class", "numpy.int64"),  # TODO: remove when possible
    ("py:class", "numpy.float32"),  # TODO: remove when possible
    ("py:class", "numpy.float64"),  # TODO: remove when possible
    ("py:class", "_io.StringIO"),
    ("py:class", "_io.BytesIO"),
    ("py:class", "pathlib._local.Path"),
    ("py:class", "pybind11_builtins.pybind11_object"),
]

autodoc_default_options = {
    "members": True,
    "undoc-members": True,
    "show-inheritance": True,
    "no-value": True,
    "special-members": "__call__",
}

always_document_param_types = True
typehints_fully_qualified = False
typehints_defaults = "braces-after"
typehints_document_rtype = True

autodoc_pydantic_model_show_json = False
autodoc_pydantic_settings_show_json = False
autodoc_pydantic_model_show_config_summary = False
autodoc_pydantic_model_signature_prefix = "schema"
autodoc_pydantic_field_doc_policy = "docstring"


def include_marked_members(_app, _what, _name, obj, skip, _options):
    doc = obj.__doc__
    if isinstance(doc, str) and ".. include in docs" in doc:
        return False
    return skip


def setup(app):
    app.connect("autodoc-skip-member", include_marked_members)



"""
Schema for stimulation protocols / source terms
-----------------------------------------------
"""

from __future__ import annotations

from typing import Any

import numpy as np
from pydantic import BaseModel, ConfigDict, Field, field_validator

from pigreads.schema.setter import Setter, normalise_list_of_setters


class SignalStep(BaseModel):
    """
    Within a time interval, increase the value of variables by a given amounts.

    The start and duration of the signal are given in time units of the
    simulation. The variables are given as a dictionary with the variable names
    as keys and the amount to increase the variable by as values::

        start: 100 # ms
        duration: 10 # ms
        variables:
          u: 1
          v: 2

    The interval is defined mathematically as
    :math:`t \\in [\\text{start}, \\text{start} + \\text{duration})`.
    """

    start: float = Field(0.0, examples=[123.4, 248.0])
    "Starting time of the signal."
    duration: float = Field(0.0, examples=[1.0, 5.0])
    "Duration of the signal."
    variables: dict[str, float] = Field(
        default_factory=dict, examples=[{"u": 10.0, "v": 0.1}, {"u": 0.1}]
    )
    "Variables to modify and the amount to modify them by."
    model_config = ConfigDict(extra="forbid", use_attribute_docstrings=True)
    "Configuration for :py:class:`pydantic.BaseModel`."

    def __call__(
        self,
        signal: np.ndarray[Any, Any],
        times: np.ndarray[Any, Any],
        varidx: dict[str, int],
    ) -> np.ndarray[Any, Any]:
        """
        Modify the signal array.

        The signal array is modified in place to increase the values of the
        variables by the given amounts within the time interval.

        :param signal: Signal array to modify.
        :param times: Times of the signal steps.
        :param varidx: Mapping of variable names to indices in the signal array,
                       see :py:meth:`pigreads.schema.simulation.Simulation.varidx`.
        :return: Modified signal array.
        """
        mask = (self.start <= times) * (times < self.start + self.duration)
        for varname, value in dict(self.variables).items():
            signal[mask, varidx[varname]] += value
        return signal


class Stimulus(BaseModel):
    """
    A stimulus is a combination of a shape and a signal.

    The shape is defined by a list of :py:class:`pigreads.schema.setter.Setter`
    objects and defines where the stimulus is applied. The signal is defined by
    a list of :py:class:`SignalStep` objects and defines how the stimulus
    changes over time.
    """

    shape: list[Setter] = Field(default_factory=list)
    "Shape of the stimulus."
    signal: list[SignalStep] = Field(default_factory=list)
    "Signal of the stimulus."
    model_config = ConfigDict(extra="forbid", use_attribute_docstrings=True)
    "Configuration for :py:class:`pydantic.BaseModel`."

    @field_validator("shape", mode="before")
    @classmethod
    def normalise_shape(cls, shape: Any) -> Any:
        """
        Normalise the setters in the shape.

        See :py:func:`pigreads.schema.setter.normalise_list_of_setters` for details.

        :param shape: Any data to try to interpret.
        :return: Normalised data.
        """

        return normalise_list_of_setters(shape)


__all__ = [
    "SignalStep",
    "Stimulus",
]



"""
Low-level implementation
------------------------
"""

from __future__ import annotations

import os
from logging import warning
from typing import Any

import numpy as np
import pyopencl as cl

from pigreads import kernels

Size = np.uint64
"""
The unsigned integer type used for sizes and indices.
"""
Int = np.int32
"""
The signed integer type used for integer arrays.
"""

StatesIdx = np.dtype(
    [
        ("Ns", Size),
        ("Nt", Size),
        ("Nz", Size),
        ("Ny", Size),
        ("Nx", Size),
        ("Nv", Size),
        ("it", Size),
        ("iz", Size),
        ("iy", Size),
        ("ix", Size),
        ("iv", Size),
    ]
)
"""
The struct used to pass array shape and indices to OpenCL kernels.
"""


def idx(
    s: Any, it: int = 0, iz: int = 0, iy: int = 0, ix: int = 0, iv: int = 0
) -> np.ndarray:
    """
    Create a ``StatesIdx`` struct to pass to OpenCL kernels.

    :param s: The array for which to create the index struct.
    :param it: Index in time.
    :param iz: Index in the z-dimension.
    :param iy: Index in the y-dimension.
    :param ix: Index in the x-dimension.
    :param iv: Index in the variable dimension.
    :return: The ``StatesIdx`` struct as a numpy array.
    """
    shape = np.ones(5, dtype=np.int64)
    shape[-len(s.shape) :] = s.shape
    return np.array(
        (
            s.dtype.itemsize,
            *shape,
            it % shape[0],
            iz % shape[1],
            iy % shape[2],
            ix % shape[3],
            iv % shape[4],
        ),
        dtype=StatesIdx,
    )


_context: cl.Context | None = None  # pylint: disable=invalid-name
"""
The OpenCL context used for computations.
"""
_queue: cl.CommandQueue | None = None  # pylint: disable=invalid-name
"""
The OpenCL command queue used for computations.
"""


class Models:
    """
    Core implementation of the ``Models`` class.

    :see: :py:class:`pigreads.models.Models` for the main interface to the models.

    :param double_precision: If ``True``, use double precision for calculations.
    :param context: OpenCL context to use for computations.
    """

    def __init__(
        self, double_precision: bool = False, context: cl.Context | None = None
    ) -> None:
        self._double_precision: bool = double_precision
        self._Nv: int = 0  # pylint: disable=invalid-name
        self._model_keys: list[str] = []
        self._model_ids: list[int] = []
        self._model_offsets: list[int] = [0]
        self._model_params: list[float] = []
        self._context: cl.Context | None = context
        self._queue: cl.CommandQueue | None = None
        self._kernels: dict[str, cl.Kernel] = {}
        self._block_size: tuple[int, int, int] = (1, 8, 8)
        self.compile(kernels.CORE)

    def prepare_context(self) -> None:
        """
        Make sure there is a context and a queue.
        """
        if self._context is None:
            global _context  # noqa: PLW0603  # pylint: disable=global-statement
            if _context is None:
                _context = cl.create_some_context()
            self._context = _context

        if self._queue is None:
            global _queue  # noqa: PLW0603  # pylint: disable=global-statement
            if _queue is None:
                _queue = cl.CommandQueue(self._context)
            self._queue = _queue

    @property
    def context(self) -> cl.Context:
        """
        The OpenCL context used for computations.
        """
        self.prepare_context()
        assert self._context is not None
        return self._context

    @property
    def queue(self) -> cl.CommandQueue:
        """
        The OpenCL command queue used for computations.
        """
        self.prepare_context()
        assert self._queue is not None
        return self._queue

    def compile(self, code: str) -> None:
        """
        Compile the given OpenCL code.

        :param code: OpenCL code to compile.
        """
        program = cl.Program(self.context, kernels.HEADER + code)

        if self.double_precision:
            for device in self.context.devices:
                assert "cl_khr_fp64" in device.extensions.split(), (
                    "The selected device does not support double precision."
                )

        cokey = "PYOPENCL_COMPILER_OUTPUT"
        coval = os.environ.get(cokey)
        os.environ[cokey] = "1"

        try:
            program.build(
                options=["-D", "DOUBLE_PRECISION"] if self.double_precision else []
            )

        except cl.CompilerWarning as e:
            warning(str(e))

        if coval is not None:
            os.environ[cokey] = coval
        else:
            del os.environ[cokey]

        for kernel in program.all_kernels():
            self._kernels[kernel.function_name] = kernel

    @property
    def double_precision(self) -> bool:
        """
        Whether double precision is used.
        """
        return self._double_precision

    @property
    def real(self) -> type:
        """
        The floating-point type used, i.e., ``np.float32`` or ``np.float64``.
        """
        return np.float64 if self.double_precision else np.float32

    @property
    def Nv(self) -> int:  # pylint: disable=invalid-name
        """
        Maximum number of state variables in the models.
        """
        return self._Nv

    def __len__(self) -> int:
        """
        Number of models.

        :return: Number of models.

        .. include in docs
        """
        return len(self._model_ids)

    def get_number_definitions(self) -> int:
        """
        Get the number of model definitions.

        :return: Number of model definitions
        """
        return len(self._model_keys)

    def get_key(self, imodel: int) -> str:
        """
        Get the key of the model with the given index.

        :param imodel: Index of the model.
        :return: Key of the model with the given index.
        """
        return self._model_keys[self._model_ids[imodel]]

    def get_parameter(self, imodel: int, iparam: int) -> float:
        """
        Get the parameter with the given indices.

        :param imodel: Index of the model.
        :param iparam: Index of the parameter.
        :return: Parameter value.
        """
        return float(self._model_params[self._model_offsets[imodel] + iparam])

    def set_parameter(self, imodel: int, iparam: int, value: float) -> None:
        """
        Set the parameter with the given indices.

        :param imodel: Index of the model.
        :param iparam: Index of the parameter.
        :param value: New parameter value.
        """
        self._model_params[self._model_offsets[imodel] + iparam] = value

    def get_block_size(self) -> tuple[int, int, int]:
        """
        Get the local work size for running OpenCL kernels.

        :return: Local work size.
        """
        return self._block_size

    def set_block_size(self, block_size: tuple[int, int, int]) -> None:
        """
        Set the local work size for running OpenCL kernels.

        :param block_size: Local work size.
        """
        assert len(block_size) == 3
        self._block_size = (int(block_size[0]), int(block_size[1]), int(block_size[2]))

    @property
    def block_size(self) -> tuple[int, int, int]:
        """
        Local work size for running OpenCL kernels.
        """
        return self.get_block_size()

    @block_size.setter
    def block_size(self, block_size: tuple[int, int, int]) -> None:
        self.set_block_size(block_size)

    def adjust_work_size(self, Nz: int, Ny: int, Nx: int) -> tuple[int, int, int]:  # pylint: disable=invalid-name
        """
        Slightly increase the global work size to be a multiple of the local
        work size.

        :param Nz: Number of points in the z-dimension.
        :param Ny: Number of points in the y-dimension.
        :param Nx: Number of points in the x-dimension.
        :return: Adjusted global work size.
        """
        bz, by, bx = self.block_size
        return (
            (Nz + bz - 1) // bz * bz,
            (Ny + by - 1) // by * by,
            (Nx + bx - 1) // bx * bx,
        )

    def add(self, key: str, code: str, Nv: int, params: np.ndarray[Any, Any]) -> None:  # pylint: disable=invalid-name
        """
        Select and enable a model with given parameters.

        :param key: The key of the model to be added.
        :param code: OpenCL code for the model.
        :param Nv: Number of variables.
        :param params: Parameter values.
        """
        if key in self._model_keys:
            self._model_ids.append(self._model_keys.index(key))

        else:
            self._model_ids.append(len(self._model_keys))
            self.compile(code)
            self._model_keys.append(key)

        self._model_offsets.append(self._model_offsets[-1] + params.size)
        self._model_params.extend(params.tolist())

        self._Nv = max(self._Nv, Nv)

    def weights(
        self,
        dz: Any,
        dy: Any,
        dx: Any,
        mask: np.ndarray[Any, Any],
        diffusivity: np.ndarray[Any, Any],
    ) -> np.ndarray[Any, Any]:
        """
        Calculate the weights for the diffusion term in the reaction-diffusion
        equation.

        :param dz: The distance between points in the z-dimension.
        :param dy: The distance between points in the y-dimension.
        :param dx: The distance between points in the x-dimension.
        :param mask: 3D boolean array encoding which points are inside the medium.
        :param diffusivity: The diffusivity matrix.
        :return: Weight matrix for the diffusion term, A 5D array of shape (1, Nz, Ny, Nx, 19).
        """

        assert mask.ndim == 5
        assert mask.shape[0] == 1
        assert mask.shape[-1] == 1
        mask_np: np.ndarray[Any, Any] = np.ascontiguousarray(mask, dtype=Int)
        _, Nz, Ny, Nx, _ = mask_np.shape  # pylint: disable=invalid-name
        mask_cl = cl.Buffer(
            self.context,
            cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR,
            hostbuf=mask_np,
        )

        while diffusivity.ndim < 5:
            diffusivity = diffusivity[np.newaxis, ...]
        assert diffusivity.ndim == 5
        assert diffusivity.shape[0] == 1
        assert diffusivity.shape[-1] == 6
        diffusivity_np: np.ndarray[Any, Any] = np.ascontiguousarray(
            diffusivity, dtype=self.real
        )
        diffusivity_cl = cl.Buffer(
            self.context,
            cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR,
            hostbuf=diffusivity_np,
        )

        weights_np: np.ndarray[Any, Any] = np.zeros(
            (1, Nz, Ny, Nx, 19), dtype=self.real
        )
        weights_cl = cl.Buffer(
            self.context,
            cl.mem_flags.WRITE_ONLY,
            weights_np.nbytes,
        )

        self._kernels["calculate_weights"](
            self.queue,
            self.adjust_work_size(Nz, Ny, Nx),
            self.block_size,
            self.real(dz),
            self.real(dy),
            self.real(dx),
            mask_cl,
            idx(mask_np),
            diffusivity_cl,
            idx(diffusivity_np),
            weights_cl,
            idx(weights_np),
        )
        cl.enqueue_copy(self.queue, weights_np, weights_cl)

        self.queue.finish()

        return weights_np

    def run(
        self,
        inhom: np.ndarray[Any, Any],
        weights: np.ndarray[Any, Any],
        states: np.ndarray[Any, Any],
        stim_signal: np.ndarray[Any, Any],
        stim_shape: np.ndarray[Any, Any],
        Nt: Any,  # pylint: disable=invalid-name
        dt: Any,
    ) -> None:
        """
        Run a Pigreads simulation.

        :param inhom: A 3D array with integer values, encoding which model to use at each point. \
                Its value is zero for points outside the medium and one or more for points inside. \
                Values larger than zero are used to select one of multiple models: \
                1 for ``models[0]``, 2 for ``models[1]``, etc.
        :param weights: The weights for the diffusion term, see :py:func:`weights`.
        :param states: The initial states of the simulation, a 4D array of shape \
                (Nz, Ny, Nx, Nv).
        :param stim_signal: A 3D array with the stimulus signal at each time point \
                for all variables, with shape (Nt, Ns, Nv).
        :param stim_shape: A 4D array specifying the shape of the stimulus, \
                with shape (Ns, Nz, Ny, Nx).
        :param Nt: The number of time steps to run the simulation for.
        :param dt: The time step size.
        """

        assert dt > 0
        assert Nt > 0

        model_ids_np: np.ndarray[Any, Any] = np.array(self._model_ids, dtype=Size)
        model_ids_cl = cl.Buffer(
            self.context,
            cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR,
            hostbuf=model_ids_np,
        )

        model_offsets_np: np.ndarray[Any, Any] = np.array(
            self._model_offsets, dtype=Size
        )
        model_offsets_cl = cl.Buffer(
            self.context,
            cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR,
            hostbuf=model_offsets_np,
        )

        model_params_np: np.ndarray[Any, Any] = np.array(
            self._model_params, dtype=self.real
        )
        model_params_cl = cl.Buffer(
            self.context,
            cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR,
            hostbuf=model_params_np,
        )

        # in/output frame
        assert states.ndim == 4
        Nz, Ny, Nx, Nv = states.shape  # pylint: disable=invalid-name
        assert Nv == self.Nv

        # even and odd frames
        states_np: np.ndarray[Any, Any] = np.full(
            (2, *states.shape), np.nan, dtype=self.real
        )
        states_np[0, ...] = states
        states_cl = cl.Buffer(
            self.context,
            cl.mem_flags.READ_WRITE | cl.mem_flags.COPY_HOST_PTR,
            hostbuf=states_np,
        )

        weights_np: np.ndarray[Any, Any] = np.ascontiguousarray(
            weights, dtype=self.real
        )
        assert weights_np.ndim == 5
        assert weights_np.shape[0] == 1
        assert weights_np.shape[-1] == 19
        weights_cl = cl.Buffer(
            self.context,
            cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR,
            hostbuf=weights_np,
        )

        inhom_np: np.ndarray[Any, Any] = np.ascontiguousarray(inhom, dtype=Int)
        assert inhom_np.ndim == 4
        assert inhom_np.shape[-1] == 1
        inhom_cl = cl.Buffer(
            self.context,
            cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR,
            hostbuf=inhom_np,
        )

        stim_signal_np: np.ndarray[Any, Any] = np.ascontiguousarray(
            stim_signal, dtype=self.real
        )
        stim_shape_np: np.ndarray[Any, Any] = np.ascontiguousarray(
            stim_shape, dtype=self.real
        )

        if stim_signal_np.size > 0 and stim_shape_np.size > 0:
            assert stim_signal_np.ndim == 5
            assert stim_signal_np.shape[2] == 1
            assert stim_signal_np.shape[3] == 1
            assert stim_signal_np.shape[-1] == Nv

            assert stim_shape_np.ndim == 5
            assert stim_shape_np.shape[0] == stim_signal_np.shape[1]
            assert stim_shape_np.shape[-1] == 1

        if stim_shape_np.size == 0:
            stim_shape_np = np.ones(1, dtype=self.real).reshape((1, 1, 1, 1, 1))

        stim_shape_cl = cl.Buffer(
            self.context,
            cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR,
            hostbuf=stim_shape_np,
        )

        work_size = self.adjust_work_size(Nz, Ny, Nx)
        set_of_model_ids = set(self._model_ids)

        self._kernels["set_outside"](
            self.queue,
            work_size,
            self.block_size,
            self.real(np.nan),
            inhom_cl,
            idx(inhom_np),
            states_cl,
            idx(states_np),
        )

        for it in range(Nt):
            idx_old = idx(states_np, it=it)
            idx_new = idx(states_np, it=it + 1)

            for mid in set_of_model_ids:
                self._kernels[f"Model_{self._model_keys[mid]}_kernel"](
                    self.queue,
                    work_size,
                    self.block_size,
                    Size(len(self)),
                    model_ids_cl,
                    model_offsets_cl,
                    model_params_cl,
                    inhom_cl,
                    idx(inhom_np),
                    weights_cl,
                    idx(weights_np),
                    states_cl,
                    idx_old,
                    states_cl,
                    idx_new,
                    self.real(dt),
                )

            if stim_signal_np.size > 0 and stim_shape_np.size > 0:
                for istim in range(stim_signal_np.shape[1]):
                    stim_shape_idx = idx(stim_shape_np, it=istim)
                    for iv in range(Nv):
                        stim_amplitude = stim_signal_np[
                            it % stim_signal_np.shape[0], istim, 0, 0, iv
                        ]
                        if abs(stim_amplitude) > 1e-10:
                            idx_stim = idx(states_np, it=it + 1, iv=iv)
                            self._kernels["add_stimulus"](
                                self.queue,
                                work_size,
                                self.block_size,
                                self.real(dt * stim_amplitude),
                                stim_shape_cl,
                                stim_shape_idx,
                                states_cl,
                                idx_stim,
                            )

        for it in range(states_np.shape[0]):
            self._kernels["set_outside"](
                self.queue,
                work_size,
                self.block_size,
                self.real(np.nan),
                inhom_cl,
                idx(inhom_np),
                states_cl,
                idx(states_np, it=it),
            )

        cl.enqueue_copy(self.queue, states_np, states_cl)

        self.queue.finish()

        states[...] = states_np[Nt % states_np.shape[0], ...]


__all__ = ["Models"]



from __future__ import annotations

import importlib.metadata
from os import linesep
from typing import Any

import numpy as np

import pigreads as pig


def test_version():
    assert importlib.metadata.version("pigreads") == pig.__version__


def test_deltas():
    z, y, x = np.meshgrid(
        [0], np.arange(0, 4, 0.2), np.linspace(0, 1, 5), indexing="ij"
    )
    dz, dy, dx = pig.deltas(z, y, x)
    assert dz == 1
    assert dy == 0.2
    assert dx == 0.25
    assert pig.delta(y, ax=-2) == dy


def test_ithildin():
    z, y, x = np.mgrid[0:1, 0:3, 0:3]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.ones_like(x, dtype=int)
    models = pig.Models()
    models.add("marcotte2017dynamical")
    models.add("aliev1996simple", k=8)
    assert models.available[models[0].key].variables == {"u": 0.0, "v": 0.0}
    states: np.ndarray[Any, Any] = models.resting_states(inhom, Nframes=2).astype(
        np.float32
    )
    weights = models.weights(dz, dy, dx, inhom, diffusivity=0)
    Nt, dt = 3, 0.01
    states[1] = models.run(inhom, weights, states[0], Nt=Nt, dt=dt)
    log, variables = pig.to_ithildin(Nt * dt, dt, dz, dy, dx, models, states, inhom)
    del log["Start date"]
    del log["Simulation parameters"]["Ithildin library version"]
    del log["Simulation parameters"]["Serial number"]
    assert log == {
        "Ithildin log version": 2,
        "Simulation parameters": {
            "Timestep dt": 0.01,
            "Frame duration": 0.03,
            "Number of frames to take": 2,
            "Name of simulation series": "pigreads",
        },
        "Geometry parameters": {
            "Number of dimensions": 3,
            "Voxel size": [1.0, 1.0, 1.0],
            "Domain size": [3, 3, 1],
        },
        "Model parameters": {
            "Model type": "Marcotte & Grigoriev 2017",
            "Class": "marcotte2017dynamical",
            "Citation": f"https://doi.org/10.1063/1.5003259{linesep}https://doi.org/10.1063/1.4915143{linesep}https://doi.org/10.1103/PhysRevLett.71.1103{linesep}https://doi.org/10.1063/1.166024",
            "Parameters": {
                "diffusivity_u": 1.0,
                "diffusivity_v": 0.05,
                "beta": 1.389,
                "eps": 0.01,
                "ustar": 1.5415,
            },
            "Initial values": {"u": 0.0, "v": 0.0},
            "Variable names": ["u", "v"],
            "Number of vars": 2,
        },
        "Model parameters 1": {
            "Model type": "Aliev & Panfilov 1996",
            "Class": "aliev1996simple",
            "Citation": "https://doi.org/10.1016/0960-0779(95)00089-5",
            "Parameters": {
                "diffusivity_u": 1.0,
                "eps0": 0.002,
                "mu1": 0.2,
                "mu2": 0.3,
                "a": 0.15,
                "k": 8.0,
            },
            "Initial values": {"u": 0.0, "v": 0.0},
            "Variable names": ["u", "v"],
            "Number of vars": 2,
        },
    }
    assert " ".join(variables.keys()) == "u v inhom"
    assert variables["u"].shape == (2, 1, 3, 3)
    assert variables["v"].shape == (2, 1, 3, 3)



"""
Diffusivity
-----------
"""

from __future__ import annotations

from typing import Any

import numpy as np

from pigreads.helper import get_upper_triangle, normalise_vector


def diffusivity_matrix(
    f: np.ndarray[Any, Any] | list[int] | tuple[int, int, int] | None = None,
    n: np.ndarray[Any, Any] | list[int] | tuple[int, int, int] | None = None,
    Df: np.ndarray[Any, Any] | float = 1.0,  # pylint: disable=invalid-name
    Ds: np.ndarray[Any, Any] | float | None = None,  # pylint: disable=invalid-name
    Dn: np.ndarray[Any, Any] | float | None = None,  # pylint: disable=invalid-name
    dtype: type = np.single,
) -> np.ndarray[Any, Any]:
    """
    Define a diffusivity matrix :math:`\\textbf D` for the reaction-diffusion equation.

    If ``f`` and ``n`` are given, the matrix is defined as:

    .. math::

        \\textbf D = \\textbf D_s \\textbf I + (\\textbf D_f - \\textbf D_s)
        \\textbf f \\textbf f^\\mathrm{T} + (\\textbf D_n - \\textbf D_s)
        \\textbf n \\textbf n^\\mathrm{T}

    :param f: The main direction of diffusion, i.e., the fibre direction. \
            3D vector over space with shape (Nz, Ny, Nx, 3). The last index \
            contains three elements: the x, y, and z component of the vector. \
            Optional if :math:`D_f=D_s=D_n`.
    :param n: The direction of weakest diffusion, i.e., the direction normal to \
            the fibre sheets. A 3D vector over space with shape (Nz, Ny, Nx, 3). The \
            last index contains three elements: the x, y, and z component of the \
            vector. Optional if :math:`D_s=D_n`.
    :param Df: The diffusivity in the direction of the fibres, :math:`\\mathbf{f}`.
    :param Ds: The diffusivity in the fibre sheets, but normal to :math:`\\mathbf{f}`. \
            If ``None``, :math:`D_s` is set to :math:`D_f`.
    :param Dn: The diffusivity in the direction normal to the fibre sheets, \
            i.e., along :math:`\\mathbf{n}`. \
            If ``None``, :math:`D_n` is set to :math:`D_s`.
    :param dtype: Data type of the arrays, i.e., single or double precision floating point numbers.
    :return: A 4D array with shape (Nz, Ny, Nx, 6).

    See also :py:func:`pigreads.helper.get_upper_triangle` for the convention used for the
    last axis of the output array.
    """

    if Ds is None:
        Ds = Df

    if Dn is None:
        Dn = Ds

    if f is None:
        assert np.allclose(Df, Ds), "If Df!=Ds, f must be given"
        f = [0, 0, 0]

    if n is None:
        assert np.allclose(Ds, Dn), "If Ds!=Dn, n must be given"
        n = [0, 0, 0]

    f = normalise_vector(f, dtype=dtype)
    n = normalise_vector(n, dtype=dtype)
    eye: np.ndarray = np.eye(3, dtype=dtype)
    eye.shape = (1, 1, 1, 3, 3)

    D = get_upper_triangle(  # pylint: disable=invalid-name
        np.einsum("...,...ij->...ij", Ds, eye)
        + np.einsum("...,...ij,...ji->...ij", Df - Ds, f, f)
        + np.einsum("...,...ij,...ji->...ij", Dn - Ds, n, n)
    )
    assert D.ndim == 4
    assert D.shape[-1] == 6
    return D


__all__ = [
    "diffusivity_matrix",
]



"""
Progress bars
-------------
"""

from __future__ import annotations

import json
import sys
import time
from collections.abc import Callable, Iterable, Iterator, Sized
from typing import Any, TextIO, TypeVar

from tqdm import tqdm

T = TypeVar("T")
"abstract type variable for generic types"


class BaseProgressIterator(Iterator[T]):
    """
    Base class for progress iterators.

    :param iterable: An iterable object.
    """

    def __init__(self, iterable: Iterable[T]) -> None:
        self._iterator = iter(iterable)

    def __iter__(self) -> Iterator[T]:
        return self

    def __next__(self) -> T:
        return next(self._iterator)


class PlainProgressIterator(BaseProgressIterator[T]):
    """
    Plain progress iterator.

    This progress indicator simply prints the current index and total number of
    items.

    :param iterable: An iterable object.
    :param file: File object to write progress to.

    :ivar total: Total number of items.
    :ivar file: File object to write progress to.
    :ivar index: The current index.
    """

    def __init__(self, iterable: Iterable[T], file: TextIO = sys.stderr) -> None:
        super().__init__(iterable)
        self.total: int | None = len(iterable) if isinstance(iterable, Sized) else None
        self.file = file
        self.index = 0

    def __next__(self) -> T:
        self.index += 1
        item = next(self._iterator)
        print(f"{self.index}/{self.total}", file=self.file)
        return item


class JSONProgressIterator(BaseProgressIterator[T]):
    """
    JSON progress iterator.

    This progress indicator prints the current index, total number of items, time
    elapsed, time remaining, and time per iteration in JSON format.

    :param iterable: An iterable object.
    :param file: File object to write progress to.

    :ivar total: Total number of items.
    :ivar file: File object to write progress to.
    :ivar index: The current index.
    :ivar start: Start time.
    """

    def __init__(self, iterable: Iterable[T], file: TextIO = sys.stderr) -> None:
        super().__init__(iterable)
        self.total: int | None = len(iterable) if isinstance(iterable, Sized) else None
        self.file = file
        self.index = 0
        self.start = time.time()

    def __next__(self) -> T:
        self.index += 1
        item = next(self._iterator)
        elapsed = time.time() - self.start
        avg = elapsed / self.index
        print(
            json.dumps(
                {
                    "iteration_index": self.index,
                    "total_iterations": self.total,
                    "time_remaining": avg * (self.total - self.index)
                    if self.total is not None
                    else None,
                    "time_elapsed": elapsed,
                    "time_per_iteration": avg,
                }
            ),
            file=self.file,
        )
        return item


PROGRESS_ITERS: dict[str, Callable[[Any], Any]] = {
    "bar": tqdm,
    "json": JSONProgressIterator,
    "plain": PlainProgressIterator,
    "none": iter,
}
"""
Mapping of progress bar types to progress indicators.

Options:

- "bar": a bar with detailed progress information
- "json": JSON progress information
- "plain": plain text progress information
- "none": no progress information
"""



"""
Pigreads Python module
----------------------

This section documents the API of Pigreads, which is the main interface for
users to interact with the library.
"""

from __future__ import annotations

from pigreads._version import version as __version__
from pigreads.compat import run, weights
from pigreads.diffusivity import diffusivity_matrix
from pigreads.helper import (
    delta,
    deltas,
    get_upper_triangle,
    normalise_vector,
    prepare_array,
    to_ithildin,
)
from pigreads.models import Models
from pigreads.schema.model import ModelDefinition

__all__ = [
    "ModelDefinition",
    "Models",
    "__version__",
    "delta",
    "deltas",
    "diffusivity_matrix",
    "get_upper_triangle",
    "normalise_vector",
    "prepare_array",
    "run",
    "to_ithildin",
    "weights",
]



"""
Interactive simulation
----------------------
"""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any

import matplotlib.pyplot as plt
import numpy as np
import yaml
from matplotlib.colors import Normalize

from pigreads.models import Models
from pigreads.schema.simulation import Simulation

if TYPE_CHECKING:
    import pygame


def get_iz_iv(
    models: Models,
    variable: str | int | None,
    index_z: int | None,
) -> tuple[int, int]:
    """
    Interpret z index and variable index from command line arguments.

    :param models: Models object.
    :param variable: Variable name or index.
    :param index_z: Index in z direction.
    :return: Tuple of z index and variable index.
    """

    iz: int = index_z or 0
    iv: int

    try:
        iv = int(variable or 0)
        assert iv >= 0, "Invalid variable index"
        assert iv < models.Nv, "Invalid variable index"

    except ValueError:
        iv = 0 if variable is None else Simulation.varidx(models)[str(variable)]

    return iz, iv


class LiveView:
    """
    Live view for a simulation.

    Plot a single variable at a fixed z index in an interactive window or save
    it to a file.

    :param models: The models used in the simulation.
    :param variable: Variable name or index.
    :param index_z: Index in z direction.
    :param dx: Grid spacing in x direction.
    :param dy: Grid spacing in y direction.
    :param vmin: Minimum value for colormap normalisation.
    :param vmax: Maximum value for colormap normalisation.
    :param cmap: The colormap to use.
    :param path: Path to save the image to. If None, an interactive window is shown.
    :param fps: Frames per second for the interactive window.
    :param click_radius: Radius of click region. If None, clicking is disabled.
    :param click_value: Value to add continuously while clicking. If None, clicking is disabled.

    :ivar models: The models used in the simulation.
    :ivar iv: Index of the variable to plot.
    :ivar iz: Index in z direction.
    :ivar dx: Grid spacing in x direction.
    :ivar dy: Grid spacing in y direction.
    :ivar vmin: Minimum value for colormap normalisation.
    :ivar vmax: Maximum value for colormap normalisation.
    :ivar cmap: The colormap to use.
    :ivar path: Path to save the image to. If None, an interactive window is shown.
    :ivar fps: Frames per second for the interactive window.
    :ivar click_radius: Radius of click region. If None, clicking is disabled.
    :ivar click_value: Value to add continuously while clicking. If None, clicking is disabled
    :ivar click_location: Current mouse click location.
    :ivar click_location_prev: Previous mouse click location.
    :ivar mouse_pressed: Whether the mouse is currently pressed.
    :ivar active: Whether the view can be updated.

    Key binds in the interactive window:

    - ``j``: Show next variable.
    - ``k``: Show previous variable.
    - ``[``: Decrease FPS.
    - ``]``: Increase FPS.

    Additional key binds when clicking is enabled:

    - ``s``: Decrease click value.
    - ``w``: Increase click value.
    - ``a``: Decrease click radius.
    - ``d``: Increase click radius.
    """

    def __init__(
        self,
        models: Models,
        variable: str | int | None = None,
        index_z: int | None = None,
        dx: float = 1.0,
        dy: float = 1.0,
        vmin: float | None = None,
        vmax: float | None = None,
        cmap: str | None = None,
        path: Path | str | None = None,
        fps: float = 60.0,
        click_radius: float | None = None,
        click_value: float | None = None,
    ) -> None:
        self.models = models
        self.iz, self.iv = get_iz_iv(models, variable, index_z)
        self.dx = dx
        self.dy = dy
        self.vmin = vmin
        self.vmax = vmax
        self.cmap = plt.get_cmap(cmap)
        self.path = path
        self.fps = fps
        self.click_radius = click_radius
        self.click_value = click_value
        self.click_location: tuple[float, float] | None = None
        self.click_location_prev: tuple[float, float] | None = None
        self.mouse_pressed = False
        self._screen: pygame.Surface | None = None
        self._clock: pygame.time.Clock | None = None
        self.active: bool = True

    @staticmethod
    def normalise_kwargs(kwargs: int | str | dict[str, Any]) -> dict[str, Any]:
        """
        Normalise command line arguments which can be given in a variety of data types.

        :param kwargs: The command line argument.
        :return: Keyword arguments for the constructor of this class.
        """

        if isinstance(kwargs, str):
            path = Path(kwargs)
            if path.is_file():
                kwargs = yaml.safe_load(path.read_bytes())
        if not isinstance(kwargs, dict):
            kwargs = {"variable": kwargs}
        assert isinstance(kwargs, dict)
        return kwargs

    def draw_thick_line(self, frame: np.ndarray[Any, Any]) -> None:
        """
        Draw a thick line between the previous and new locations with the given radius.

        :param frame: Frame to modify.
        """
        if (
            self.click_radius is None
            or self.click_value is None
            or self.click_location is None
        ):
            return

        x1, y1 = self.click_location
        x2, y2 = self.click_location

        if self.click_location_prev is not None:
            x2, y2 = self.click_location_prev

        dx, dy = x2 - x1, y2 - y1

        y, x = np.meshgrid(
            self.dy * np.arange(frame.shape[-2]),
            self.dx * np.arange(frame.shape[-1]),
            indexing="ij",
        )

        with np.errstate(divide="ignore", invalid="ignore"):
            distance = np.abs(+dy * x - dx * y + x2 * y1 - y2 * x1) / np.linalg.norm(
                (dy, dx), axis=0
            )
            t = ((x - x1) * dx + (y - y1) * dy) / (dx**2 + dy**2)

        mask = (distance <= self.click_radius) & (t >= 0) & (t <= 1)

        for y_, x_ in [(y1, x1), (y2, x2)]:
            mask[np.linalg.norm([y - y_, x - x_], axis=0) <= self.click_radius] = True

        frame[mask] = self.click_value
        self.click_location_prev = self.click_location

    def update(self, states: np.ndarray[Any, Any], time: float) -> bool:
        """
        Update the live view.

        :param states: Array of states with shape (Nz, Ny, Nx, Nv).
        :param time: Simulated time.
        :return: True if the update was completed, false if the update was skipped.
        """
        if not self.active:
            return False

        self.iz = self.iz % states.shape[0]
        self.iv = self.iv % states.shape[-1]
        frame = states[self.iz, :, :, self.iv]
        Ny, Nx = frame.shape  # pylint: disable=invalid-name
        norm = Normalize(
            vmin=np.nanmin(frame) if self.vmin is None else self.vmin,
            vmax=np.nanmax(frame) if self.vmax is None else self.vmax,
        )
        frame_rgb = (self.cmap(norm(frame.T))[..., :3] * 255).astype(np.uint8)

        if self.path is not None:
            plt.imsave(self.path, np.transpose(frame_rgb, (1, 0, 2)))

        else:
            import pygame  # pylint: disable=import-outside-toplevel

            timer = pygame.USEREVENT + 1  # pylint: disable=no-member

            if self._screen is None:
                pygame.init()  # pylint: disable=no-member
                self._screen = pygame.display.set_mode(
                    (Nx, Ny),
                    pygame.RESIZABLE,  # pylint: disable=no-member
                    vsync=1,
                )
                self._clock = pygame.time.Clock()
                pygame.time.set_timer(timer, 100)  # every 100 ms
                pygame.event.post(pygame.event.Event(timer))
            assert self._screen is not None
            assert self._clock is not None

            wx, wy = self._screen.get_size()
            scale = min(wx / Nx, wy / Ny)
            sx, sy = scale * Nx, scale * Ny
            ox, oy = int((wx - sx) / 2 + 0.5), int((wy - sy) / 2 + 0.5)

            if pygame.mouse.get_pressed()[0]:
                mx, my = pygame.mouse.get_pos()
                mx = (mx - ox) / sx * Nx * self.dx
                my = (my - oy) / sy * Ny * self.dy
                self.click_location = mx, my
                self.mouse_pressed = True

            else:
                self.mouse_pressed = False
                self.click_location = None
                self.click_location_prev = None

            self.draw_thick_line(frame)

            surface = pygame.surfarray.make_surface(frame_rgb)
            surface = pygame.transform.smoothscale(
                surface, (int(sx + 0.5), int(sy + 0.5))
            )

            self._screen.fill((0, 0, 0))
            self._screen.blit(surface, (ox, oy))
            pygame.display.flip()
            self._clock.tick(self.fps)

            for event in pygame.event.get():
                if event.type == timer:
                    pygame.display.set_caption(
                        f"Pigreads — iv={self.iv} "
                        + (f"iz={self.iz} " if states.shape[0] > 1 else "")
                        + f"time={time:.3f} "
                        + f"fps={self._clock.get_fps():.0f}:{self.fps:.0f}"
                    )
                elif event.type == pygame.QUIT:  # pylint: disable=no-member
                    pygame.quit()  # pylint: disable=no-member
                    self.active = False

                elif event.type == pygame.KEYDOWN:  # pylint: disable=no-member
                    match event.key:
                        case pygame.K_LEFTBRACKET:  # pylint: disable=no-member
                            self.fps /= 1.2
                        case pygame.K_RIGHTBRACKET:  # pylint: disable=no-member
                            self.fps *= 1.2
                        case pygame.K_k:  # pylint: disable=no-member
                            self.iv -= 1
                        case pygame.K_j:  # pylint: disable=no-member
                            self.iv += 1
                        case pygame.K_a:  # pylint: disable=no-member
                            if self.click_radius is not None:
                                self.click_radius /= 1.2
                        case pygame.K_d:  # pylint: disable=no-member
                            if self.click_radius is not None:
                                self.click_radius *= 1.2
                        case pygame.K_s:  # pylint: disable=no-member
                            if self.click_value is not None:
                                self.click_value /= 1.2
                        case pygame.K_w:  # pylint: disable=no-member
                            if self.click_value is not None:
                                self.click_value *= 1.2

        return self.active


__all__ = [
    "LiveView",
]



"""
Command Line Interface
----------------------
"""

from __future__ import annotations

import json
import re
from importlib.util import find_spec
from pathlib import Path
from typing import Any, TextIO
from warnings import warn

import click
import numpy as np
import yaml

import pigreads.plot as pp
from pigreads import Models, __version__
from pigreads.gui import LiveView, get_iz_iv
from pigreads.progress import PROGRESS_ITERS
from pigreads.schema.model import ModelDefinition
from pigreads.schema.simulation import Simulation

_cmc_available: bool = False
if find_spec("cmcrameri") is not None:
    import cmcrameri  # type: ignore[import-untyped]

    assert cmcrameri
    _cmc_available = True


@click.group()
@click.option(
    "-m",
    "--define-model",
    "model_defs",
    type=click.File("r"),
    multiple=True,
    help="Define a model using a YAML file. "
    "The beginning of the filename will be used as the model key.",
)
def main(model_defs: list[TextIO]) -> None:
    "Run and visualise Pigreads simulations."

    for model_def in model_defs:
        stem = Path(model_def.name).stem
        match = re.match(r"^[0-9A-Za-z_]+", stem)
        message = f"Filename '{stem}' must start with an alphanumeric identifier!"
        assert match, message
        key = match.group(0)
        message = f"Defining Models.available['{key}'] using {model_def.name}..."
        warn(message, RuntimeWarning, stacklevel=2)
        Models.available[key] = ModelDefinition(**yaml.safe_load(model_def))


@main.command()
def version() -> None:
    "Show installed version."
    click.echo(f"Pigreads {__version__}")


@main.command()
@click.argument("config", type=click.File("r"))
@click.argument("result", type=click.Path())
@click.option(
    "-p",
    "--progress",
    type=click.Choice(list(PROGRESS_ITERS.keys())),
    default="bar",
    show_default=True,
    help="Progress bar type.",
)
@click.option(
    "-s", "--start-frame", type=int, default=0, help="Index of frame to start with."
)
@click.option(
    "-l",
    "--live",
    type=yaml.safe_load,
    default="{}",
    show_default=True,
    help="YAML encoded keyword arguments for LiveView to show a live preview of "
    "a chosen variable. Instead of a YAML string, a path to a YAML file may "
    "also be given, or the name or index of a variable.",
)
@click.option(
    "-i",
    "--infinite",
    is_flag=True,
    default=False,
    help="Run the simulation in an infinite loop.",
)
def run(
    config: TextIO,
    result: click.Path,
    progress: str,
    start_frame: int,
    live: int | str | dict[str, Any],
    infinite: bool,
) -> None:
    """
    Run a simulation.

    CONFIG is a YAML or JSON file defining the simulation,
    and RESULT is a path to write the results to.
    """

    sim = Simulation(**yaml.safe_load(config))
    live_: dict[str, Any] = LiveView.normalise_kwargs(live)
    if len(live_) > 0 and "cmap" not in live_:
        live_["cmap"] = "cmc.acton" if _cmc_available else None
    liveview = (
        LiveView(models=sim.prepare_models(), dy=sim.dy, dx=sim.dx, **live_)
        if len(live_) > 0
        else None
    )

    def callback(states: np.ndarray[Any, Any], ifr: int) -> bool:
        return (
            True
            if liveview is None
            else liveview.update(states[ifr], time=sim.Nt * sim.dt * ifr)
        )

    sim.run(
        path=str(result),
        progress=PROGRESS_ITERS[progress],
        start_frame=start_frame,
        callback=callback,
        infinite=infinite,
    )


@main.command()
@click.argument("input", type=click.File("r"), default="-")
@click.argument("output", type=click.File("w"), default="-")
@click.option(
    "-f",
    "--format",
    type=click.Choice(["yaml", "json"]),
    default="yaml",
    show_default=True,
    help="Output format.",
)
def convert(
    input: TextIO,  # pylint: disable=redefined-builtin
    output: TextIO,
    format: str,  # pylint: disable=redefined-builtin
) -> None:
    """
    Convert a configuration file to canonical format.

    INPUT is a YAML or JSON file defining the simulation,
    and OUTPUT is a path to write the converted file to.
    """
    sim = Simulation(**yaml.safe_load(input))
    data = sim.model_dump()

    if format == "yaml":
        yaml.safe_dump(data, output, allow_unicode=True)
    else:
        assert format == "json"
        json.dump(data, output, indent=2)


@main.command()
@click.argument("config", type=click.File("r"))
@click.argument("result", type=click.Path(exists=True))
@click.argument("movie", type=click.Path())
@click.option(
    "-d", "--dpi", type=int, default=180, show_default=True, help="Dots per inch."
)
@click.option(
    "-r", "--fps", type=int, default=15, show_default=True, help="Frames per second."
)
@click.option(
    "-p",
    "--progress",
    type=click.Choice(list(PROGRESS_ITERS.keys())),
    default="bar",
    show_default=True,
    help="Progress bar type.",
)
@click.option("-z", "--index-z", type=int, default=None, help="Index in z direction.")
@click.option(
    "-v", "--variable", type=str, default=None, help="Variable name or index."
)
@click.option(
    "--kwargs",
    type=yaml.safe_load,
    default="{}",
    show_default=True,
    help="Additional arguments as YAML or JSON to pass to movie, "
    "plot_frame, and imshow, for example "
    '\'{"cmap": "inferno", "vmin": -0.25, "vmax": 1.25}\'.',
)
@click.option(
    "-n",
    "--parallel",
    type=int,
    default=1,
    show_default=True,
    help="Number of processes. 0 to use all CPUs.",
)
def movie(
    config: TextIO,
    result: click.Path,
    movie: click.Path,  # pylint: disable=redefined-outer-name
    dpi: int,
    fps: int,
    progress: str,
    variable: str | None,
    index_z: int | None,
    parallel: int,
    kwargs: dict[str, Any],
) -> None:
    """
    Generate a movie from a result file using FFmpeg and Matplotlib.

    CONFIG is a YAML or JSON file defining the simulation,
    RESULT is a NPY file containing the simulation results,
    and MOVIE is a path to write an MP4 video to.
    """

    sim = Simulation(**yaml.safe_load(config))
    models = sim.prepare_models()

    array: np.ndarray[Any, Any] = np.lib.format.open_memmap(
        str(result),
        mode="r",
    )
    assert array.shape == (sim.Nfr, sim.Nz, sim.Ny, sim.Nx, models.Nv)
    iz, iv = get_iz_iv(models, variable, index_z or sim.Nz // 2)
    array = array[:, iz, :, :, iv]

    zlabel = f" z = {sim.dz * iz:.2f}" if sim.Nz > 1 else ""
    tlables = [
        f"t = {time}" + zlabel for time in (sim.dt * sim.Nt * np.arange(sim.Nfr))
    ]

    pp.movie(
        path=str(movie),
        frames=array,
        tlables=tlables,
        progress=progress,
        parallel=parallel,
        **pp.imshow_defaults(
            sim=sim,
            array=array,
            dpi=dpi,
            fps=fps,
            vlabel=list(models.available[models[0].key].variables)[iv],
            **kwargs,
        ),
    )


__all__ = [
    "convert",
    "main",
    "movie",
    "run",
    "version",
]



#!/bin/env python3
"""
Run the Niederer benchmark using Pigreads.

https://doi.org/10.1098/rsta.2011.0139
"""

from __future__ import annotations

import warnings

import matplotlib.pyplot as plt
import numpy as np
from tqdm import tqdm

import pigreads as pig
from pigreads.plot import movie


def main():
    framedur = 1.0  # ms
    dt = 0.01  # ms {0.005, 0.01, 0.05}
    Nt = int(framedur / dt + 0.5)

    dz = dy = dx = 0.2  # mm {0.1, 0.2, 0.5}
    Lz, Ly, Lx = 3, 7, 20  # mm
    z, y, x = np.meshgrid(
        *[
            d * np.arange(1 + round(L / d))
            for d, L in zip([dz, dy, dx], [Lz, Ly, Lx], strict=True)
        ],
        indexing="ij",
    )
    assert dz == z[1, 0, 0]
    assert dy == y[0, 1, 0]
    assert dx == x[0, 0, 1]
    assert Lz == z[-1, 0, 0]
    assert Ly == y[0, -1, 0]
    assert Lx == x[0, 0, -1]

    inhom = np.ones_like(x, dtype=int)
    inhom[-1] = 0
    inhom[:, -1] = 0
    inhom[:, :, -1] = 0

    # constants
    C_m = 0.01  # µF/mm² (membrane capacitance)
    chi = 140  # 1/mm (surface-to-volume ratio)

    # diffusivities intra/extra longitudinal/transversal
    scale = 1 / (chi * C_m)  # mm³/µF
    Dif = scale * 0.17  # mm²/ms
    Dis = scale * 0.019  # mm²/ms
    Def = scale * 0.62  # mm²/ms
    Des = scale * 0.24  # mm²/ms

    # diffusivities longitudinal/transversal
    Df = Dif * Def / (Dif + Def)  # mm²/ms
    Ds = Dis * Des / (Dis + Des)  # mm²/ms

    # stimulus current
    I_stim = 50.0  # mA/cm³ (stimulus current)
    i_stim = I_stim * scale  # mV/ms

    models = pig.Models()
    models.add("tentusscher2006alternans", diffusivity_V=Df)
    diffusivity = pig.diffusivity_matrix(Df=1, Ds=Ds / Df, f=[1, 0, 0])
    weights = models.weights(dz, dy, dx, inhom, diffusivity)

    states = models.resting_states(inhom, Nframes=100)

    stim_shape = (x < 1.5) & (y < 1.5) & (z < 1.5)
    stim_signal = np.zeros((Nt, states.shape[-1]))
    stim_signal[:, 0] = i_stim

    for ifr in tqdm(range(states.shape[0] - 1)):
        states[ifr + 1] = models.run(
            inhom,
            weights,
            states[ifr],
            stim_signal=stim_signal if ifr <= 2 else None,
            stim_shape=stim_shape if ifr <= 2 else None,
            Nt=Nt,
            dt=dt,
        )

        if ifr > 0 and not np.all(
            np.isfinite(states[ifr + 1][np.isfinite(states[ifr])])
        ):
            with warnings.catch_warnings():
                warnings.simplefilter("always")
                warnings.warn("New NaN values found!", stacklevel=2)

    # plot data with Matplotlib
    plt.imshow(states[-1, states.shape[1] // 2, :, :, 0])
    plt.colorbar()
    plt.show()

    # create a movie using FFmpeg
    movie(
        "benchmark.mp4",
        states[:, states.shape[1] // 2, :-1, :-1, 0],
        progress="bar",
        parallel=1,
        tlables=[f"frame {ifr}" for ifr in range(states.shape[0])],
        extent=(x[0, 0, 0], x[0, 0, -2], y[0, 0, 0], y[0, -2, 0]),
        vmin=-100,
        vmax=30,
    )


if __name__ == "__main__":
    main()



"""
Schema to define simulations
----------------------------

This submodule provides a Pydantic schema defining a simulation. The schema is
used to read and write simulation configuration files in JSON or YAML format.

See :py:class:`pigreads.schema.simulation.Simulation` for the main schema.
"""

from __future__ import annotations

from pigreads.schema.basic import Slice, Vector3D
from pigreads.schema.diffusivity import Diffusivity, DiffusivityFile, DiffusivityParams
from pigreads.schema.model import ModelDefinition, ModelEntry
from pigreads.schema.setter import Setter, SetterFile
from pigreads.schema.simulation import Simulation
from pigreads.schema.stimulus import SignalStep, Stimulus

__all__ = [
    "Diffusivity",
    "DiffusivityFile",
    "DiffusivityParams",
    "ModelDefinition",
    "ModelEntry",
    "Setter",
    "SetterFile",
    "SignalStep",
    "Simulation",
    "Slice",
    "Stimulus",
    "Vector3D",
]



"""
Schema for model definitions
----------------------------
"""

from __future__ import annotations

from collections.abc import MutableMapping
from typing import Any

from pydantic import BaseModel, ConfigDict, Field, model_validator


class ModelEntry(BaseModel):
    """
    A model entry chooses a model by its key and which parameters to use.

    :see: :py:class:`pigreads.models.Models` and :py:class:`ModelDefinition`
    """

    key: str = Field(..., examples=["author2025title", "aliev1996simple", "trivial"])
    "The key of the model."
    parameters: MutableMapping[str, float] = Field(
        default_factory=dict, examples=[{"a": 0.2, "diffusivity_u": 1.2}, {}]
    )
    "The parameters of the model."
    model_config = ConfigDict(extra="forbid", use_attribute_docstrings=True)
    "Configuration for :py:class:`pydantic.BaseModel`."

    @model_validator(mode="before")
    @classmethod
    def normalise(cls, values: Any) -> Any:
        """
        Normalise the model entry.

        If the entry is a string, it is converted to a dictionary with an empty
        parameter set. If the entry is a dictionary with a single key, the key
        is used as the model key.

        :param values: Any data to try to interpret.
        :return: Normalised data.
        """

        assert isinstance(values, dict)

        key = values.get("key")
        if "key" in values:
            del values["key"]

        parameters = values.get("parameters", {})
        if "parameters" in values:
            del values["parameters"]

        if key is None:
            assert len(values) == 1

            key, parameters = next(iter(values.items()))
            return {"key": key, "parameters": parameters or {}}

        parameters = {**parameters, **values}

        return {"key": key, "parameters": parameters or {}}


class ModelDefinition(BaseModel):
    """
    Definition of a so-called model, i.e., the reaction-term.

    Note: Additional fields are allowed stored in :py:attr:`meta`.

    :see: :py:class:`pigreads.models.Models` and :py:class:`ModelEntry`
    """

    name: str = Field(..., examples=["author2025title", "aliev1996simple", "trivial"])
    "Human readable name of the model, usually the authors and the year."
    description: str = Field(
        ..., examples=["This two-variable model was created by modifying the model..."]
    )
    "A simple description of the model, what it was designed for, and its unique features."
    dois: list[str] = Field(
        default_factory=list,
        examples=[
            ["https://doi.org/10.0000/foobar0123"],
            [
                "https://doi.org/10.1016/j.jtbi.2008.03.029",
                "https://doi.org/10.1063/1.166311",
                "https://doi.org/10.1161/01.RES.82.11.1206",
                "https://doi.org/10.1152/ajpheart.00794.2003",
            ],
        ],
    )
    "List of digital object identifiers in URL form, i.e., starting with ``https://doi.org/10``."
    variables: dict[str, float] = Field(
        default_factory=dict, examples=[{"u": 0.0, "v": 1.0}]
    )
    "Model variable names and their resting/initial values."
    diffusivity: dict[str, float] = Field(default_factory=dict, examples=[{"u": 1.0}])
    "Dictionary that maps the names of variables to diffuse to their diffusivity."
    parameters: dict[str, float] = Field(
        default_factory=dict,
        examples=[{"eps0": 0.002, "mu1": 0.2, "mu2": 0.3, "a": 0.15, "k": 8.0}],
    )
    "Model parameter names and their default values."
    code: str = Field(..., examples=["*_new_u = u + dt * _diffuse_u;"])
    "Source code of the forward Euler step in OpenCL."
    model_config = ConfigDict(extra="allow", use_attribute_docstrings=True)
    "Configuration for :py:class:`pydantic.BaseModel`."

    @property
    def meta(self) -> dict[str, Any]:
        """
        Read-only view of additional fields, i.e., metadata.
        """
        assert self.model_extra is not None
        return self.model_extra

    @property
    def all_parameters(self) -> dict[str, float]:
        "Dictionary of parameters including the diffusivities."
        return {
            **{
                "diffusivity_" + varname: value
                for varname, value in self.diffusivity.items()
            },
            **self.parameters,
        }

    def __call__(self, **parameters: float) -> dict[str, float]:
        """
        Merge diffusivities, default parameters, and given parameters.

        :param parameters: Additional parameters.
        :return: Merged parameter dictionary.
        """
        merged = {**self.all_parameters}
        for parameter, value in parameters.items():
            error = f"Unknown parameter: {parameter}"
            assert parameter in merged, error
            merged[parameter] = value
        return merged


__all__ = [
    "ModelDefinition",
    "ModelEntry",
]



"""
Setters for arrays
------------------
"""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal

import numpy as np
from pydantic import BaseModel, ConfigDict, Field, TypeAdapter, model_validator

from pigreads.schema import basic

if TYPE_CHECKING:
    import pigreads
    import pigreads.schema


class SetterFile(BaseModel):
    """
    Set an array to the contents of a file.

    The file should contain a numpy array with the same shape as the array to be set.

    :see: :py:func:`numpy.load`
    """

    cmd: Literal["file"] = "file"
    "Identifier for this setter."
    file: Path = Field(..., examples=[Path("path/to/file.npy")])
    "Path to the file containing the array."
    model_config = ConfigDict(extra="forbid", use_attribute_docstrings=True)
    "Configuration for :py:class:`pydantic.BaseModel`."

    def __call__(
        self, array: np.ndarray[Any, Any], _: pigreads.schema.Simulation
    ) -> np.ndarray[Any, Any]:
        """
        Set the array to the contents of the file.

        :param array: Array to set.
        :return: Array set to the contents of the file.
        """
        array[:] = np.load(self.file)
        return array


class SetterSlices(BaseModel):
    """
    Set slices of an array to a value.

    The slices are defined by a list of :py:class:`pigreads.schema.basic.Slice` objects.
    Only where all slices are true, the value is set.
    """

    cmd: Literal["slice"] | Literal["slices"] = "slices"
    "Identifier for this setter."
    value: float = Field(..., examples=[0.123, 0.0, 1.0, np.nan])
    "Value to set in the slices."
    slices: list[basic.Slice] = Field(default_factory=list)
    "List of slices."
    model_config = ConfigDict(extra="forbid", use_attribute_docstrings=True)
    "Configuration for :py:class:`pydantic.BaseModel`."

    def __call__(self, array: np.ndarray[Any, Any], _: Any) -> np.ndarray[Any, Any]:
        """
        Set the slices of the array to the value.

        :param array: Array to set.
        :return: Array with the slices set to the value.
        """
        array = np.array(array)
        selection: list[int | slice] = [slice(None)] * array.ndim
        for i, _ in enumerate(selection):
            for s in self.slices:
                j = s.axis
                if j < 0:
                    j += array.ndim
                if i == j:
                    selection[i] = s()
        array[tuple(selection)] = self.value
        return array

    @model_validator(mode="before")
    @classmethod
    def normalise(cls, values: Any) -> Any:
        """
        Normalise the setter.

        If there is only one slice, it is converted to a list of slices.

        :param values: Any data to try to interpret.
        :return: Normalised data.
        """

        if isinstance(values, dict):
            axis = values.get("axis")
            start = values.get("start")
            end = values.get("end")
            step = values.get("step")
            if axis is not None:
                values["slices"] = values.get("slices", [])
                values["slices"].append(
                    {
                        "axis": axis,
                        "start": start,
                        "end": end,
                        "step": step,
                    }
                )
                for key in ["axis", "start", "end", "step"]:
                    if key in values:
                        del values[key]

        return values


class SetterSpherical(BaseModel):
    """
    Set an array to a value inside or outside a sphere or ellipsoid.

    The ellipsoid is defined by a center and a radii in each dimension. If not
    none, the inside and outside values are set to given values.

    If not given, the radius is set to half the size of the array in each dimension,
    and the center is set to the center of the array.
    """

    cmd: Literal["spherical"] = "spherical"
    "Identifier for this setter."
    outside: float | None = Field(None, examples=[0.123, 0.0, 1.0, np.nan])
    "Value to set outside the ellipsoid, ``None`` for no change."
    inside: float | None = Field(None, examples=[0.123, 0.0, 1.0, np.nan])
    "Value to set inside the ellipsoid, ``None`` for no change."
    radius: basic.Vector3D | float | None = Field(
        None, examples=[1.0, basic.Vector3D(x=1.0, y=0.5, z=0.1)]
    )
    "Radius of the sphere or radii of the ellipsoid, ``None`` for half the size of the array."
    center: basic.Vector3D | None = Field(
        None, examples=[basic.Vector3D(x=4.0, y=3.0, z=0.0)]
    )
    "Center of the ellipsoid, ``None`` for the center of the array."
    exponent: basic.Vector3D | float = Field(
        2, examples=[2.0, 1.0, basic.Vector3D(x=2.0, y=2.0, z=1.0)]
    )
    "Exponent for the norm, usually 2 for a sphere or ellipsoid."
    model_config = ConfigDict(extra="forbid", use_attribute_docstrings=True)
    "Configuration for :py:class:`pydantic.BaseModel`."

    def __call__(
        self, array: np.ndarray[Any, Any], sim: pigreads.schema.Simulation
    ) -> np.ndarray[Any, Any]:
        """
        Modify the array to set values inside or outside a sphere or ellipsoid.

        The array is modified in place to set the values inside or outside the
        sphere or ellipsoid to the given values.

        :param array: Array to modify.
        :param sim: Simulation object with the grid information,
                    see :py:class:`pigreads.schema.simulation.Simulation`.
        :return: Modified array.
        """

        assert array.ndim == 3

        deltas: list[float] = [sim.dz, sim.dy, sim.dx]
        shape: list[float] = list(array.shape)

        radius: list[float]
        if isinstance(self.radius, basic.Vector3D):
            radius = [self.radius.z, self.radius.y, self.radius.x]
            radius = [np.inf if r <= 0 else r for r in radius]
        elif self.radius is None:
            radius = [
                dx * Nx / 2 if Nx > 1 else np.inf
                for dx, Nx in zip(deltas, shape, strict=True)
            ]
        else:
            radius = [self.radius, self.radius, self.radius]

        center: list[float]
        if isinstance(self.center, basic.Vector3D):
            center = [self.center.z, self.center.y, self.center.x]
        else:
            assert self.center is None
            center = [dx * Nx / 2 for dx, Nx in zip(deltas, shape, strict=True)]

        exponent: list[float]
        if isinstance(self.exponent, basic.Vector3D):
            exponent = [self.exponent.z, self.exponent.y, self.exponent.x]
        else:
            exponent = [self.exponent, self.exponent, self.exponent]

        mask = (
            np.sum(
                np.meshgrid(
                    *(
                        np.abs(
                            (np.array([x0]) if Nx <= 1 else dx * np.arange(Nx) - x0)
                            / Rx
                        )
                        ** ex
                        for dx, Nx, x0, Rx, ex in zip(
                            deltas, shape, center, radius, exponent, strict=True
                        )
                    ),
                    indexing="ij",
                ),
                axis=0,
            )
            < 1
        )

        if self.inside is not None:
            array[mask] = self.inside

        if self.outside is not None:
            array[~mask] = self.outside

        return array


Setter = SetterFile | SetterSlices | SetterSpherical
"""
A setter is a command to set an array to a given value or to modify the array
in a specific way.

One of:

- :py:class:`SetterFile`
- :py:class:`SetterSlices`
- :py:class:`SetterSpherical`
"""


def normalise_list_of_setters(values: Any) -> Any:
    """
    Normalise a list of setters.

    Instead of the full representation::

        setters:
        - cmd: spherical
          outside: 0

    The command string can also be given as the only key of a dictionary::

        setters:
        - spherical:
            outside: 0

    Also instead of a list, a dictionary can be given with the command as key::

        setters:
          spherical:
            outside: 0

    If there is a single string given, it is interpreted as a file setter::

        setters: "file.npy"

    If a setter is given as a string::

        setters:
          file: "file.npy"

    It is converted to a dictionary::

        setters:
          file:
            file: "file.npy"

    This function is used in the schema to normalise the input
    and to allow more flexible input formats.

    :param values: Any data to try to interpret.
    :return: Normalised data.
    """

    adapter: TypeAdapter[Setter] = TypeAdapter(Setter)

    if isinstance(values, dict):
        values = [
            {"cmd": k, **({} if v is None else ({k: v} if isinstance(v, str) else v))}
            for k, v in values.items()
        ]
    elif isinstance(values, str) or isinstance(values, Path):  # noqa: SIM101  # pylint: disable=consider-merging-isinstance
        values = [SetterFile(file=Path(values))]
    elif isinstance(values, list):
        values_: list[Any] = []
        for i, _ in enumerate(values):
            obj = values[i]
            if isinstance(obj, dict):
                cmd = obj.get("cmd")
                if "cmd" in obj:
                    del obj["cmd"]
                if cmd is None and len(obj) == 1:
                    cmd, obj = next(iter(obj.items()))
                if isinstance(obj, str):
                    obj = {cmd: obj}
                obj = {"cmd": cmd, **(obj or {})}
            values_.append(obj)
        values = [adapter.validate_python(cmd) for cmd in values_]
    return values


__all__ = [
    "Setter",
    "SetterFile",
    "SetterSlices",
    "SetterSpherical",
    "normalise_list_of_setters",
]



"""
Schema for a simulation
-----------------------
"""

from __future__ import annotations

from collections.abc import Callable
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal
from warnings import warn

import numpy as np
from pydantic import BaseModel, ConfigDict, Field, field_validator

from pigreads.schema.diffusivity import Diffusivity
from pigreads.schema.model import ModelEntry
from pigreads.schema.setter import Setter, SetterFile, normalise_list_of_setters
from pigreads.schema.stimulus import Stimulus

if TYPE_CHECKING:
    from pigreads import Models

# pylint: disable=import-outside-toplevel


class Simulation(BaseModel):
    """
    Definition of a simulation.

    The simulation is defined on a three dimensional cuboid with :py:attr:`Nz`,
    :py:attr:`Ny`, and :py:attr:`Nx` grid points in the z, y, and x directions
    respectively. The grid points are separated by :py:attr:`dz`, :py:attr:`dy`,
    and :py:attr:`dx` units in the z, y, and x directions. The simulation is
    run for :py:attr:`Nfr` frames with :py:attr:`Nt` time steps per frame.
    Each time step is :py:attr:`dt` units long.

    The simulation definion contains a list of
    :py:class:`pigreads.schema.model.ModelEntry` objects
    (:py:attr:`models`) which define the reaction terms in the
    reaction-diffusion equation. See also :py:class:`pigreads.models.Models`.

    The diffusivity matrix is defined by
    a :py:class:`pigreads.schema.diffusivity.Diffusivity` object.

    The inhomogeneity (:py:attr:`inhom`) of the medium, i.e., which parts of the
    simulation are active and which model to use where, is defined by a list of
    :py:class:`pigreads.schema.setter.Setter` objects.

    The initial state of the simulation is defined by a dictionary of
    :py:class:`pigreads.schema.setter.Setter` objects. The keys of the
    dictionary are the variable names and the values are lists of setters.

    A stimulus protocol is defined by a list
    of :py:class:`pigreads.schema.stimulus.Stimulus` objects.

    The flag :py:attr:`double_precision` can be set to `True` to use double
    precision floating point numbers in the simulation.

    The required key :py:attr:`pigreads` is used to identify the simulation
    file format.

    The simulation can be run by calling the :py:meth:`run` method.

    Example::

        pigreads: 1
        Nfr: 100
        Nt: 200
        Nz: 1
        Ny: 200
        Nx: 200
        dt: 0.025
        dz: 1
        dy: 0.1
        dx: 0.1
        inhom:
        - spherical:
            outside: 0
        diffusivity: 0.03
        models:
        - marcotte2017dynamical:
            diffusivity_u: 1.0
            diffusivity_v: 0.05
        init:
            u:
                slices:
                    value: 1
                    axis: -1
                    end: 10
            v:
                slices:
                    value: 2
                    start: 100
                    axis: -2

    Read this and run a simulation like this::

        from pigreads.schema import Simulation
        data = yaml.safe_load(yaml_string)
        sim = Simulation(**data)
        sim.run(path="result.npy")

    Or use the CLI:

        $ pigreads run config.yaml result.npy

    """

    pigreads: Literal[1]
    "Identifier for the simulation file format."

    Nfr: int = Field(..., ge=1, examples=[100])
    "Number of frames."
    Nt: int = Field(..., ge=1, examples=[200])
    "Number of time steps per frame."
    Nz: int = Field(..., ge=1, examples=[1, 200])
    "Number of grid points in z direction."
    Ny: int = Field(..., ge=1, examples=[200, 1])
    "Number of grid points in y direction."
    Nx: int = Field(..., ge=1, examples=[200, 1])
    "Number of grid points in x direction."

    dt: float = Field(..., gt=0, examples=[0.025])
    "Time step."
    dz: float = Field(..., gt=0, examples=[1.0, 0.1])
    "Grid spacing in z direction."
    dy: float = Field(..., gt=0, examples=[0.1, 1.0])
    "Grid spacing in y direction."
    dx: float = Field(..., gt=0, examples=[0.1, 1.0])
    "Grid spacing in x direction."

    models: list[ModelEntry]
    """
    Definition of the models to use in the simulation.

    The models define the reaction terms in the reaction-diffusion equation.

    :see: :py:class:`pigreads.models.Models`, :py:meth:`prepare_models`
    """

    diffusivity: Diffusivity
    """
    Definition of the diffusivity matrix.

    :see: :py:func:`pigreads.diffusivity.diffusivity_matrix`
    """

    inhom: list[Setter] = Field(default_factory=list)
    """
    Definition of the inhomogeneity of the medium.

    It defines which parts of the simulation are active and which model to
    use where.

    :see: :py:class:`pigreads.schema.setter.Setter`, :py:meth:`prepare_inhom`,
          :py:func:`pigreads.models.Models.weights`
    """

    init: dict[str, list[Setter]] | SetterFile = Field(
        default_factory=dict,  # type: ignore[arg-type]
    )
    """
    Definition of the initial state of the simulation.

    The initial state is defined by a dictionary of variable names and lists
    of setters.

    :see: :py:class:`pigreads.schema.setter.Setter`, :py:meth:`prepare_states`
    """

    stim: list[Stimulus] = Field(default_factory=list)
    """
    Definition of the stimulus protocol.

    The stimulus protocol is defined by a list
    of :py:class:`pigreads.schema.stimulus.Stimulus` objects.

    :see: :py:class:`pigreads.schema.stimulus.Stimulus`, :py:meth:`prepare_stim`
    """

    double_precision: bool = Field(False, examples=[False, True])
    "Flag to use single or double precision floating point numbers."

    model_config = ConfigDict(extra="forbid", use_attribute_docstrings=True)
    "Configuration for :py:class:`pydantic.BaseModel`."

    @staticmethod
    def is_dict_of_dicts(
        d: dict[Any, Any], excluded_keys: list[Any] | None = None
    ) -> bool:
        """
        Check if a value of a dictionary is also a dictionary.

        :param d: Dictionary to check.
        :param excluded_keys: Ignore values whose key is in excluded_keys.
        :return: True if it is a dictionary of dictionaries.
        """
        return next(
            (
                True
                for k, v in d.items()
                if k not in (excluded_keys or []) and isinstance(v, dict)
            ),
            False,
        )

    @field_validator("models", mode="before")
    @classmethod
    def normalise_models(cls, models: Any) -> Any:
        """
        Normalise the models.

        Normally, the models are given as a list of
        :py:class:`pigreads.schema.model.ModelEntry`.

        Alternatively, if the models are given as a string, it is converted to
        a list with a single model entry. If the models are given as a
        dictionary, it is converted to a list of model entries. If the models
        are given as a list of strings, they are converted to a list of model
        entries with empty parameter sets.

        :param models: Any data to try to interpret.
        :return: Normalised data.
        """
        if models is None:
            return models
        if isinstance(models, dict):
            if cls.is_dict_of_dicts(models, excluded_keys=["key", "parameters"]):
                models = [
                    ModelEntry(key=k, parameters=v or {}) for k, v in models.items()
                ]
            else:
                models = [ModelEntry(**models)]
        elif isinstance(models, str):
            models = [ModelEntry(key=models, parameters={})]
        elif isinstance(models, list):
            models = [
                ModelEntry(**model)
                if isinstance(model, dict)
                else ModelEntry(key=model, parameters={})
                for model in models
            ]
        return models

    @field_validator("diffusivity", mode="before")
    @classmethod
    def normalise_diffusivity(cls, diffusivity: Any) -> Any:
        """
        Normalise the diffusivity.

        Normally, the diffusivity is given as
        an instance of :py:class:`pigreads.schema.diffusivity.Diffusivity`.

        Alternatively, if the diffusivity is given as a string, read the
        diffusivity matrix from the file. If the diffusivity is given as a
        float, set the diffusivity matrix to a constant value.

        :param diffusivity: Any data to try to interpret.
        :return: Normalised data.
        """

        if isinstance(diffusivity, str):
            diffusivity = {"file": diffusivity}
        elif isinstance(diffusivity, float) or isinstance(diffusivity, int):  # noqa: SIM101  # pylint: disable=consider-merging-isinstance
            diffusivity = {"Df": diffusivity}
        return diffusivity

    @field_validator("inhom", mode="before")
    @classmethod
    def normalise_inhom(cls, inhom: Any) -> Any:
        """
        Normalise the inhomogeneity.

        The inhomogeneity is given as lists of
        :py:class:`pigreads.schema.setter.Setter` objects.

        The setters are normalised using
        :py:func:`pigreads.schema.setter.normalise_list_of_setters`.

        :param inhom: Any data to try to interpret.
        :return: Normalised data.
        """
        return normalise_list_of_setters(inhom)

    @field_validator("init", mode="before")
    @classmethod
    def normalise_init(cls, init: Any) -> Any:
        """
        Normalise the initial state.

        The initial state is given as lists of
        :py:class:`pigreads.schema.setter.Setter` objects.

        The setters are normalised using
        :py:func:`pigreads.schema.setter.normalise_list_of_setters`.

        :param init: Any data to try to interpret.
        :return: Normalised data.
        """
        if init is not None:
            if isinstance(init, dict):
                for varname, setters in init.items():
                    if isinstance(setters, dict):
                        init[varname] = normalise_list_of_setters(setters)
                    elif isinstance(setters, float) or isinstance(setters, int):  # noqa: SIM101  # pylint: disable=consider-merging-isinstance
                        init[varname] = normalise_list_of_setters(
                            {"slices": {"axis": -1, "value": setters}}
                        )
                    elif isinstance(setters, str):
                        init[varname] = normalise_list_of_setters({"file": setters})
            else:
                assert isinstance(init, str) or isinstance(init, Path)  # noqa: SIM101  # pylint: disable=consider-merging-isinstance
                return SetterFile(file=Path(init))

        return init

    @staticmethod
    def varidx(models: Models) -> dict[str, int]:
        """
        Create a dictionary mapping variable names to indices.

        The indices are used to access the variables in the state array.

        :param models: Models object.
        :return: Mapping of variable names to indices.
        """
        return {
            varname: i
            for i, varname in enumerate(models.available[models[0].key].variables)
        }

    def prepare_space(
        self,
    ) -> tuple[np.ndarray[Any, Any], np.ndarray[Any, Any], np.ndarray[Any, Any]]:
        """
        Get arrays with the coordinates of the grid points.

        The arrays are returned as a tuple of three arrays with the z, y, and x
        coordinates of the grid points.

        :return: Arrays for the spatial coordinates z, y, and x.
        :see: :py:func:`numpy.meshgrid`
        """
        z, y, x = np.meshgrid(
            self.dz * np.arange(self.Nz),
            self.dy * np.arange(self.Ny),
            self.dx * np.arange(self.Nx),
            indexing="ij",
        )
        return z, y, x

    def prepare_inhom(self) -> np.ndarray[Any, Any]:
        """
        Get the inhomogeneity array.

        The inhomogeneity array is created by applying the setters in the
        :py:attr:`inhom` list to an array of ones with the shape of the grid.

        The inhomogeneity array is used to define the active parts of the
        simulation and which model to use where.

        :return: Inhomogeneity array.
        """

        inhom: np.ndarray[Any, Any] = np.ones((self.Nz, self.Ny, self.Nx), dtype=int)
        for setter in self.inhom:
            inhom = setter(inhom, self)
        return inhom

    def prepare_models(self) -> Models:
        """
        Create a models object from the model entries.

        The model entries are used to create the models object which is used to
        run the simulation.

        :return: Models object.
        """

        from pigreads import Models

        models = Models(double_precision=self.double_precision)
        for model in self.models:
            models.add(model.key, **model.parameters)
        return models

    def prepare_states(
        self,
        models: Models,
        varidx: dict[str, int],
        inhom: np.ndarray[Any, Any],
        path: Path | str | None = None,
    ) -> np.ndarray[Any, Any]:
        """
        Prepare the state array.

        The state array is created as a five dimensional array with the shape
        ``(Nfr, Nz, Ny, Nx, Nv)`` where ``Nv`` is the number of variables in the
        model, see also :py:func:`pigreads.models.Models.Nv`.

        This array is created in three steps:

        1. Create a new array with the correct shape and dtype using
           :py:meth:`pigreads.helper.prepare_array`.
        2. Set the first frame of the array to the resting state of the model,
           see :py:func:`pigreads.models.Models.resting_states`.
        3. Apply the initial state setters to the first frame of the array.

        The array is returned as a memory map if a path is given, otherwise as a
        normal numpy array.

        :param models: Models object.
        :param varidx: Mapping of variable names to indices in the state array,
                       see :py:func:`Simulation.varidx`.
        :param inhom: Inhomogeneity array.
        :param path: Path to memory map the state array to.
        :return: State array.
        """
        from pigreads import prepare_array

        states = prepare_array(
            shape=(self.Nfr, self.Nz, self.Ny, self.Nx, len(varidx)),
            path=path,
            dtype=np.float32,
        )
        states[0] = models.resting_states(inhom, Nframes=1)

        if isinstance(self.init, SetterFile):
            states[0] = self.init(states[0], self)  # pylint: disable=not-callable

        else:
            assert isinstance(self.init, dict)
            for varname, setters in dict(self.init).items():
                for setter in setters:
                    idx = varidx[varname]
                    states[0, ..., idx] = setter(states[0, ..., idx], self)

        return states

    def prepare_stim(
        self, varidx: dict[str, int]
    ) -> tuple[np.ndarray[Any, Any], np.ndarray[Any, Any]]:
        """
        Prepare the stimulus signal and shape.

        The stimulus signal is created as a three dimensional array with the
        shape ``((Nfr - 1)*Nt, Ns, Nv)`` where ``Ns`` is the number of
        stimuli in the simulation and ``Nv`` is the number of variables in the
        model, see also :py:func:`pigreads.models.Models.Nv`.
        This array is first set to zero and then modified by the signal steps
        in the stimulus objects,
        see :py:class:`pigreads.schema.stimulus.SignalStep`
        and :py:class:`pigreads.schema.stimulus.Stimulus`.

        The stimulus shape is created as a four dimensional array with the
        shape ``(Ns, Nz, Ny, Nx)`` where ``Ns`` is the number of stimuli in the
        simulation. This array is first set to zero and then modified by the
        shape setters in the stimulus objects,
        see :py:class:`pigreads.schema.setter.Setter`
        and :py:class:`pigreads.schema.stimulus.Stimulus`.

        The stimulus signal and shape are returned as a tuple.

        :param varidx: Mapping of variable names to indices in the state array
                       see :py:func:`Simulation.varidx`.
        :return: Stimulus signal and shape.
        """

        Nt: int = (self.Nfr - 1) * self.Nt  # pylint: disable=invalid-name
        Ns: int = len(self.stim)  # pylint: disable=invalid-name
        Nv: int = len(varidx)  # pylint: disable=invalid-name

        stim_signal = np.zeros((Nt, Ns, Nv))
        times = self.dt * np.arange(Nt)
        for i, stimulus in enumerate(self.stim):
            for step in stimulus.signal:
                stim_signal[:, i, :] = step(stim_signal[:, i, :], times, varidx)

        stim_shape = np.zeros((Ns, self.Nz, self.Ny, self.Nx), dtype=float)
        for i, stimulus in enumerate(self.stim):
            for setter in stimulus.shape:
                stim_shape[i] = setter(stim_shape[i], self)

        return stim_signal, stim_shape

    def run(
        self,
        path: Path | str | None = None,
        start_frame: int = 0,
        progress: Callable[[Any], Any] = lambda x: x,
        callback: Callable[[np.ndarray[Any, Any], int], bool] | None = None,
        infinite: bool = False,
    ) -> np.ndarray[Any, Any]:
        """
        Run the simulation.

        :param start_frame: The frame to start the simulation at, default is 0.
        :param path: The path to save the state array to, default is ``None``,
                     see :py:meth:`pigreads.helper.prepare_array`.
        :param progress: A function to show progress, default is no progress updates.
        :param callback: A function to call after each frame, default is no callback.
                         If it returns ``False``, stop iteration early.
        :param infinite: Whether to run the simulation in an infinite loop, default is ``False``.
        :return: The results in the state array with the shape ``(Nfr, Nz, Ny, Nx, Nv)``.

        The simulation calls the following methods to prepare the simulation:

        - :py:meth:`prepare_inhom`
        - :py:meth:`prepare_models`
        - :py:meth:`prepare_states`
        - :py:meth:`prepare_stim`
        - :py:meth:`pigreads.models.Models.weights`

        The simulation is then run frame by frame using the
        :py:func:`pigreads.models.Models.run` function.
        The results are saved to the state array and returned.
        """

        inhom = self.prepare_inhom()
        models = self.prepare_models()
        varidx = self.varidx(models)
        states = self.prepare_states(models, varidx, inhom, path=path)
        stim_signal, stim_shape = self.prepare_stim(varidx)
        weights = models.weights(
            self.dz, self.dy, self.dx, inhom, diffusivity=self.diffusivity()
        )

        if callback is not None:
            callback(states, 0)

        Nfr: int = states.shape[0]  # pylint: disable=invalid-name

        def step(old: int) -> bool:
            active: bool = True
            new: int = (old + 1) % Nfr
            states[new] = models.run(
                inhom,
                weights,
                states[old],
                stim_signal[old * self.Nt : new * self.Nt],
                stim_shape,
                Nt=self.Nt,
                dt=self.dt,
            )

            if callback is not None:
                active = callback(states, new)

            if old > 0 and np.any(np.isfinite(states[old]) > np.isfinite(states[new])):
                warn(
                    f"New non-finite values found at frame {new}!",
                    RuntimeWarning,
                    stacklevel=2,
                )

            return active

        if infinite:
            ifr = start_frame
            while step(ifr):
                ifr = (ifr + 1) % Nfr

        else:
            for ifr in progress(range(start_frame, states.shape[0] - 1)):
                if not step(ifr):
                    break

        return states


__all__ = [
    "Simulation",
]



#!/bin/env python3
"""
Run a simulation.
"""

from __future__ import annotations

import warnings

import matplotlib.pyplot as plt
import numpy as np
from tqdm import tqdm

import pigreads as pig
from pigreads.plot import movie


def main():
    random = np.random.default_rng(seed=0)

    R = 11.05
    z, y, x = np.mgrid[0:1, -R:R:200j, -R:R:200j]
    dz, dy, dx = pig.deltas(z, y, x)
    r = np.linalg.norm((x, y, z), axis=0)

    inhom = np.ones_like(x, dtype=int)
    inhom[r > R] = 0
    inhom *= random.uniform(0, 1, size=x.shape) > 0.1

    models = pig.Models()
    models.add("marcotte2017dynamical")

    states = models.resting_states(inhom, Nframes=100)
    states[0, np.linalg.norm(((x + 8), y, z), axis=0) < 2, 0] = 1
    states[0, y < 0, 1] = 2

    diffusivity = pig.diffusivity_matrix(Df=0.03)

    weights = models.weights(dz, dy, dx, inhom, diffusivity)

    Nt = 200
    dt = 0.025
    for ifr in tqdm(range(states.shape[0] - 1)):
        states[ifr + 1] = models.run(inhom, weights, states[ifr], Nt=Nt, dt=dt)

        if ifr > 0 and not np.all(
            np.isfinite(states[ifr + 1][np.isfinite(states[ifr])])
        ):
            with warnings.catch_warnings():
                warnings.simplefilter("always")
                warnings.warn("New NaN values found!", stacklevel=2)

    # plot data with Matplotlib
    plt.imshow(
        states[-1, 0, :, :, 0],
        extent=(x[0, 0, 0], x[0, 0, -1], y[0, 0, 0], y[0, -1, 0]),
    )
    plt.show()

    # create a movie using FFmpeg
    movie(
        "minimal.mp4",
        states[:, 0, :, :, 0],
        progress="bar",
        tlables=[f"frame {ifr}" for ifr in range(states.shape[0])],
        extent=(x[0, 0, 0], x[0, 0, -1], y[0, 0, 0], y[0, -1, 0]),
    )


if __name__ == "__main__":
    main()



#!/bin/env python3
"""
Define a model and run a simulation.
"""

from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np
from tqdm import tqdm

import pigreads as pig
from pigreads.schema.model import ModelDefinition

pig.Models.available["fitzhugh1961impulses"] = ModelDefinition(
    name="FitzHugh 1961 & Nagumo 1962",
    description="A 2D simplification of the Hodgkin-Huxley model.",
    dois=[
        "https://doi.org/10.1016/S0006-3495(61)86902-6",
        "https://doi.org/10.1109/JRPROC.1962.288235",
    ],
    variables={"u": 1.2, "v": -0.625},
    diffusivity={"u": 1.0},
    parameters={"a": 0.7, "b": 0.8, "c": 3.0, "z": 0.0},
    code="""
        *_new_u = u + dt * (v + u - u*u*u/3 + z + _diffuse_u);
        *_new_v = v + dt * (-(u - a + b*v)/c);
    """,
)


def main():
    R = 10
    z, y, x = np.mgrid[0:1, 0:1, -R:R:200j]
    dz, dy, dx = pig.deltas(z, y, x)
    r = np.linalg.norm((x, y, z), axis=0)

    inhom = np.ones_like(x, dtype=int)
    inhom[r >= R] = 0

    models = pig.Models()
    models.add("fitzhugh1961impulses", z=-0.4)

    states = models.resting_states(inhom, Nframes=100)
    states[0, x < -8, 0] = -1

    diffusivity = pig.diffusivity_matrix(Df=1.0)

    weights = pig.weights(dz, dy, dx, inhom, diffusivity)

    Nt = 100
    dt = 0.004
    for it in tqdm(range(states.shape[0] - 1)):
        states[it + 1] = pig.run(models, inhom, weights, states[it], Nt=Nt, dt=dt)

        if it > 0:
            assert np.all(np.isfinite(states[it + 1][np.isfinite(states[it])])), (
                "New NaN values found!"
            )

    plt.imshow(
        states[:, 0, 0, :, 0].T,
        extent=(0, (states.shape[0] - 1) * Nt * dt, x[0, 0, 0], x[0, 0, -1]),
        origin="lower",
    )
    plt.colorbar()
    plt.show()


if __name__ == "__main__":
    main()



"""
Backwards compatibility
-----------------------

While these functions are deprecated, they are kept in the API for legacy
reasons.
"""

from __future__ import annotations

from typing import Any
from warnings import warn

import numpy as np

from pigreads.models import Models


def weights(
    dz: float = 1.0,
    dy: float = 1.0,
    dx: float = 1.0,
    inhom: np.ndarray[Any, Any] | None = None,
    diffusivity: np.ndarray[Any, Any] | float = 1.0,
    double_precision: bool = False,
) -> np.ndarray[Any, Any]:
    """
    Calculate the weights for the diffusion term in the reaction-diffusion
    equation.

    .. deprecated:: 0.5.0
       Use :func:`pigreads.models.Models.weights` instead.

    :param dz: The distance between points in the z-dimension,\
            see :py:func:`pigreads.helper.deltas`.
    :param dy: The distance between points in the y-dimension.
    :param dx: The distance between points in the x-dimension.
    :param inhom: A 3D array with integer values, encoding which model to use at each point. \
            Its value is zero for points outside the medium and one or more for points inside. \
            If ``None``, all points are considered inside the medium.
    :param diffusivity: The diffusivity matrix,\
            see :py:func:`pigreads.diffusivity.diffusivity_matrix`. \
            If a scalar is given, the matrix is isotropic with the same value in all directions.
    :param double_precision: If ``True``, use double precision for calculations.
    :return: Weight matrix for the diffusion term, A 5D array of shape (1, Nz, Ny, Nx, 19).
    """
    message: str = (
        "This function is deprecated! Use pigreads.models.Models.weights instead."
    )
    warn(message, DeprecationWarning, stacklevel=2)
    models = Models(double_precision=double_precision)
    return models.weights(dz, dy, dx, inhom=inhom, diffusivity=diffusivity)


def run(
    models: Models,
    inhom: np.ndarray[Any, Any],
    weights: np.ndarray[Any, Any],  # pylint: disable=redefined-outer-name
    states: np.ndarray[Any, Any],
    stim_signal: np.ndarray[Any, Any] | None = None,
    stim_shape: np.ndarray[Any, Any] | None = None,
    Nt: int = 1,  # pylint: disable=invalid-name
    dt: float = 0.001,
    double_precision: bool | None = None,
) -> np.ndarray[Any, Any]:
    """
    Run a Pigreads simulation.

    .. deprecated:: 0.5.0
       Use :func:`pigreads.models.Models.run` instead.

    :param models: The models to be used in the simulation,\
            see :py:class:`pigreads.models.Models`.
    :param inhom: A 3D array with integer values, encoding which model to use at each point. \
            Its value is zero for points outside the medium and one or more for points inside. \
            Values larger than zero are used to select one of multiple models: \
            1 for ``models[0]``, 2 for ``models[1]``, etc.
    :param weights: The weights for the diffusion term, see :py:func:`weights`.
    :param states: The initial states of the simulation, a 4D array of shape \
            (Nz, Ny, Nx, Nv), see :py:func:`pigreads.models.Models.resting_states`.
    :param stim_signal: A 3D array with the stimulus signal at each time point \
            for all variables, with shape (Nt, Ns, Nv). If ``None``, no stimulus is applied.
    :param stim_shape: A 4D array specifying the shape of the stimulus, \
            with shape (Ns, Nz, Ny, Nx). If ``None``, no stimulus is applied
    :param Nt: The number of time steps to run the simulation for.
    :param dt: The time step size.
    :param double_precision: If ``True``, use double precision for calculations.
    :return: The final states of the simulation, a 4D array of shape (Nz, Ny, Nx, Nv).
    """
    message: str = (
        "This function is deprecated! Use pigreads.models.Models.run instead."
    )
    warn(message, DeprecationWarning, stacklevel=2)
    if double_precision is not None:
        assert models.double_precision == double_precision, (
            "Chosen precision must match with instance of Models."
        )
    return models.run(
        inhom=inhom,
        weights=weights,
        states=states,
        stim_signal=stim_signal,
        stim_shape=stim_shape,
        Nt=Nt,
        dt=dt,
    )


__all__ = [
    "run",
    "weights",
]



from __future__ import annotations

import json
import re
import warnings
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest import mock

import numpy as np
import pytest
import yaml
from click.testing import CliRunner

from pigreads import Models, __version__
from pigreads.cli import main as cli
from pigreads.progress import PROGRESS_ITERS, BaseProgressIterator
from pigreads.schema import Simulation

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import pygame

yaml_data = """
pigreads: 1
Nfr: 3
Nt: 2
Nz: 1
Ny: 1
Nx: 1
dt: 0.1
dz: 0.1
dy: 0.1
dx: 0.1
diffusivity: 0.1
models: aliev1996simple
init:
    u: 0.3
"""
sim = Simulation(**yaml.safe_load(yaml_data))

model_data = """
name: Test
description: A test model
dois: []
variables:
  u: 0
diffusivity:
  u: 1.0
parameters:
  a: 0.001
code: |
  *_new_u = u + dt * (_diffuse_u - a*u);
"""


def test_usage():
    runner = CliRunner()
    result = runner.invoke(cli)
    print(result.output)
    assert result.exit_code in [0, 2]
    assert result.output.startswith("Usage")


def test_version():
    runner = CliRunner()
    result = runner.invoke(cli, ["version"])
    print(result.output)
    assert result.exit_code == 0
    assert result.stdout.strip() == f"Pigreads {__version__}"


def test_define_model():
    with TemporaryDirectory() as tempdir:
        path = Path(tempdir)
        path_model = path / "test.yaml"

        with path_model.open("w") as f:
            f.write(model_data)

        runner = CliRunner()
        with pytest.warns(RuntimeWarning, match="Defining Models.available"):
            result = runner.invoke(
                cli,
                ["-m", str(path_model), "version"],
                catch_exceptions=False,
            )
        assert result.exit_code == 0
        assert "test" in Models.available
        assert Models.available["test"].name == "Test"


def test_yaml2json():
    with TemporaryDirectory() as tempdir:
        path = Path(tempdir)
        path_input = path / "input.yaml"
        path_output = path / "output.json"

        with path_input.open("w") as f:
            f.write(yaml_data)

        runner = CliRunner()
        result = runner.invoke(
            cli,
            ["convert", "-f", "json", str(path_input), str(path_output)],
        )
        assert result.exit_code == 0

        with path_output.open() as f:
            assert Simulation(**json.loads(f.read())) == sim


def test_json2yaml():
    with TemporaryDirectory() as tempdir:
        path = Path(tempdir)
        path_input = path / "input.json"
        path_output = path / "output.yaml"

        with path_input.open("w") as f:
            f.write(sim.model_dump_json())

        runner = CliRunner()
        result = runner.invoke(
            cli,
            ["convert", str(path_input), str(path_output)],
        )
        assert result.exit_code == 0

        with path_output.open() as f:
            assert Simulation(**yaml.safe_load(f.read())) == sim


def test_invalid_conversion():
    with TemporaryDirectory() as tempdir:
        path = Path(tempdir)
        path_input = path / "input.json"
        path_output = path / "output.yaml"

        with path_input.open("w") as f:
            f.write(sim.model_dump_json())

        runner = CliRunner()
        result = runner.invoke(
            cli,
            [
                "convert",
                "-f",
                "invalid",
                str(path_input),
                str(path_output),
            ],
        )
        assert result.exit_code == 2
        assert "Invalid value" in result.output


def test_run():
    with TemporaryDirectory() as tempdir:
        path = Path(tempdir)
        path_input = path / "input.yaml"
        path_output = path / "output.npy"

        with path_input.open("w") as f:
            f.write(yaml_data)

        runner = CliRunner()
        result = runner.invoke(cli, ["run", str(path_input), str(path_output)])
        assert result.exit_code == 0

        states = np.load(path_output)
        assert np.allclose(states[-1].squeeze(), [0.4401616, 0.00194737])


def test_live():
    with TemporaryDirectory() as tempdir:
        path = Path(tempdir)
        path_input = path / "input.yaml"
        path_output = path / "output.npy"
        path_live = path / "live.png"

        with path_input.open("w") as f:
            f.write(yaml_data)

        runner = CliRunner()
        result = runner.invoke(
            cli,
            [
                "run",
                str(path_input),
                str(path_output),
                "--live",
                f"{{variable: u, path: {path_live}}}",
            ],
        )
        assert result.exit_code == 0

        states = np.load(path_output)
        assert np.allclose(states[-1].squeeze(), [0.4401616, 0.00194737])


def test_live_quit():
    with TemporaryDirectory() as tempdir:
        path = Path(tempdir)
        path_input = path / "input.yaml"
        path_output = path / "output.npy"
        path_live = path / "live.png"

        with path_input.open("w") as f:
            f.write(yaml_data)

        runner = CliRunner()
        with (
            mock.patch(
                "pygame.event", mock.Mock(get=lambda: [mock.Mock(type=pygame.QUIT)])
            ),
            mock.patch("pygame.quit"),
        ):
            result = runner.invoke(
                cli,
                [
                    "run",
                    str(path_input),
                    str(path_output),
                    "--live",
                    f"{{variable: u, path: {path_live}}}",
                ],
            )
        assert result.exit_code == 0


def test_progress():
    with TemporaryDirectory() as tempdir:
        path = Path(tempdir)
        path_input = path / "input.yaml"
        path_output = path / "output.npy"

        with path_input.open("w") as f:
            f.write(yaml_data)

        for progress in PROGRESS_ITERS:
            runner = CliRunner()
            result = runner.invoke(
                cli, ["run", "-p", progress, str(path_input), str(path_output)]
            )
            assert result.exit_code == 0


def test_progress_base():
    for _ in BaseProgressIterator(range(3)):
        pass


def test_movie():
    with TemporaryDirectory() as tempdir:
        path = Path(tempdir)
        path_input = path / "input.yaml"
        path_output = path / "output.npy"
        path_movie = path / "movie.mp4"

        with path_input.open("w") as f:
            f.write(yaml_data)

        runner = CliRunner()
        result = runner.invoke(cli, ["run", str(path_input), str(path_output)])
        assert result.exit_code == 0

        result = runner.invoke(
            cli, ["movie", str(path_input), str(path_output), str(path_movie)]
        )
        assert result.exit_code == 0


def test_unstable():
    with TemporaryDirectory() as tempdir:
        path = Path(tempdir)
        path_input = path / "input.yaml"
        path_output = path / "output.npy"

        with path_input.open("w") as f:
            f.write(re.sub(r"dt:.*", "dt: 1e10", yaml_data))

        runner = CliRunner()
        with pytest.warns(RuntimeWarning):
            result = runner.invoke(cli, ["run", str(path_input), str(path_output)])
        assert result.exit_code == 0



from __future__ import annotations

import os
from typing import Any

import numpy as np
import pytest

import pigreads as pig
from pigreads.schema.model import ModelDefinition


def test_inhom():
    z, y, x = np.mgrid[0:1, 0:1, 0:3]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.arange(3).reshape((1, 1, 3))
    models = pig.Models()
    models.add("aliev1996simple")
    models.add("gray1983autocatalytic")
    states = models.resting_states(inhom, Nframes=2)
    states[0, 0, 0, 0, :] = 1.0  # outside
    states[0, 0, 0, 1, 0] = 0.4  # aliev1996simple u
    states[0, 0, 0, 2, 1] = 0.5  # gray1983autocatalytic v
    weights = models.weights(dz, dy, dx, inhom, diffusivity=0)
    states[1] = models.run(inhom, weights, states[0], Nt=1, dt=0.1)
    assert np.allclose(states[0, 0, 0, 0, :], 1), "starting at given outside value"
    assert np.all(np.isnan(states[1, 0, 0, 0, :])), "outside value must be set to nan"
    assert states[0, 0, 0, 1, 0] < states[1, 0, 0, 1, 0], (
        "aliev1996simple u must increase"
    )
    assert states[0, 0, 0, 1, 1] < states[1, 0, 0, 1, 1], (
        "aliev1996simple v must increase"
    )
    assert states[0, 0, 0, 2, 0] > states[1, 0, 0, 2, 0], (
        "gray1983autocatalytic u must decrease"
    )
    assert states[0, 0, 0, 2, 1] < states[1, 0, 0, 2, 1], (
        "gray1983autocatalytic v must increase"
    )


def test_stim():
    z, y, x = np.mgrid[0:1, 0:1, 0:5]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.ones_like(x)
    inhom[:, :, -1] = 2
    models = pig.Models()
    models.add("trivial")
    models.add("aliev1996simple", k=0, eps0=0, mu1=0)
    states = models.resting_states(inhom, Nframes=2)
    weights = models.weights(dz, dy, dx, inhom, diffusivity=0)
    stim_signal = np.array([(0, 0), (-1, 2), (4, 1), (np.nan, np.nan)])
    stim_shape = np.array([0, 1, 0.5, -1, 1]).reshape(inhom.shape)
    dt = 1.0
    states[1] = models.run(
        inhom,
        weights,
        states[0],
        stim_signal=stim_signal,
        stim_shape=stim_shape,
        Nt=3,
        dt=dt,
    )
    assert states.shape == (2, 1, 1, 5, 2)
    sum_u, sum_v = np.sum(stim_signal[:3], axis=0)
    for i in range(4):
        assert states[1, 0, 0, i, 0] == dt * stim_shape[0, 0, i] * sum_u, f"i={i}"
    assert states[1, 0, 0, 4, 1] == dt * stim_shape[0, 0, 4] * sum_v


def test_stims():
    z, y, x = np.mgrid[0:1, 0:1, 0:5]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.ones_like(x)
    inhom[:, :, -1] = 2
    models = pig.Models()
    models.add("trivial")
    models.add("aliev1996simple", k=0, eps0=0, mu1=0)
    states = models.resting_states(inhom, Nframes=2)
    weights = models.weights(dz, dy, dx, inhom, diffusivity=0)
    stim_signal = np.array(
        [[[6, 2], [0, 0]], [[8, 9], [6, 7]], [[5, 9], [8, 0]], [[8, 0], [7, 1]]]
    )
    stim_shape = np.array([[[[8, 5, 2, 4, 0]]], [[[1, 6, 6, 6, 3]]]])
    dt = 1.0
    states[1] = models.run(
        inhom,
        weights,
        states[0],
        stim_signal=stim_signal,
        stim_shape=stim_shape,
        Nt=3,
        dt=dt,
    )
    assert states.shape == (2, 1, 1, 5, 2)
    s = states[1].squeeze()
    assert np.allclose(s[:, 0], [166, 179, 122, 160, -336])
    assert np.all(np.isnan(s[:4, 1]))
    assert s[4, 1] == 21


def test_stim_dt():
    z, y, x = np.mgrid[0:1, 0:1, 0:1]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.ones_like(x)
    models = pig.Models("trivial")
    states = models.resting_states(inhom, Nframes=3)
    weights = models.weights(dz, dy, dx, inhom, diffusivity=0)
    stim_signal = np.arange(5).reshape((-1, 1))
    stim_shape = inhom
    dt = 0.1
    states[1] = dt * models.run(
        inhom,
        weights,
        states[0],
        stim_signal=stim_signal,
        stim_shape=stim_shape,
        Nt=10,
        dt=1,
    )
    states[2] = models.run(
        inhom,
        weights,
        states[0],
        stim_signal=stim_signal,
        stim_shape=stim_shape,
        Nt=10,
        dt=dt,
    )
    assert np.allclose(states[1], states[2])


def test_cell():
    z, y, x = np.mgrid[0:1, 0:1, 0:1]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.ones_like(x)
    models = pig.Models("aliev1996simple")
    models.core.block_size = (1, 1, 1)
    states = models.resting_states(inhom, Nframes=3)
    states[0, ..., 0] = 0.4
    weights = models.weights(dz, dy, dx, inhom, diffusivity=0)
    for i in range(states.shape[0] - 1):
        states[i + 1] = models.run(inhom, weights, states[i], Nt=199, dt=0.1)
    assert states.shape == (3, 1, 1, 1, 2)
    states = states.squeeze()
    assert np.allclose(
        states,
        [
            [4.00000006e-01, 0.00000000e00],
            [9.06229615e-01, 5.90469241e-01],
            [4.87683921e-11, 1.08949475e-01],
        ],
    )


def test_deprecatred():
    z, y, x = np.mgrid[0:1, 0:1, 0:1]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.ones_like(x)
    models = pig.Models("aliev1996simple")
    models.core.block_size = (1, 1, 1)
    states = models.resting_states(inhom, Nframes=3)
    states[0, ..., 0] = 0.4
    with pytest.warns(DeprecationWarning):
        weights = pig.weights(dz, dy, dx, inhom, diffusivity=0)
    for i in range(states.shape[0] - 1):
        with pytest.warns(DeprecationWarning):
            states[i + 1] = pig.run(models, inhom, weights, states[i], Nt=199, dt=0.1)
    assert states.shape == (3, 1, 1, 1, 2)
    states = states.squeeze()
    assert np.allclose(
        states,
        [
            [4.00000006e-01, 0.00000000e00],
            [9.06229615e-01, 5.90469241e-01],
            [4.87683921e-11, 1.08949475e-01],
        ],
    )


def test_cable():
    z, y, x = np.mgrid[0:1, 0:1, 0:50]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.ones_like(x)
    inhom[..., -1] = 0
    models = pig.Models("aliev1996simple")
    states = models.resting_states(inhom, Nframes=3)
    states[0, x < 5, 0] = 0.4
    weights = models.weights(dz, dy, dx, inhom, diffusivity=1)
    Nt, dt = 399, 0.05
    for i in range(states.shape[0] - 1):
        states[i + 1] = models.run(inhom, weights, states[i], Nt=Nt, dt=dt)
    assert states.shape == (3, 1, 1, 50, 2)
    assert np.allclose(
        states[:, 0, 0, (0, 25, 48), 0],
        [
            [4.0000001e-01, 0.0000000e00, 0.0000000e00],
            [9.0334904e-01, 9.9558353e-01, 1.0182223e-14],
            [1.0085634e-08, 5.3530955e-01, 9.9503601e-01],
        ],
    )


def test_spiral():
    z, y, x = np.mgrid[0:1, -25:25, -25:25]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.ones_like(x)
    inhom[:, -1, :] = 0
    inhom[:, :, -1] = 0
    models = pig.Models("aliev1996simple")
    states = models.resting_states(inhom, Nframes=3)
    states[0, x > 0, 0] = 1
    states[0, y > 0, 1] = 1
    weights = models.weights(dz, dy, dx, inhom, diffusivity=1)
    Nt, dt = 399, 0.05
    for i in range(states.shape[0] - 1):
        states[i + 1] = models.run(inhom, weights, states[i], Nt=Nt, dt=dt)
    indices = 0, 10, 40, 48
    assert states[0, 0, 49, 49, 0] == 1.0
    assert np.all(np.isnan(states[1:, 0, 49, 49, :])), (
        "outside value must be set to nan"
    )
    assert np.allclose(
        states[:, 0, indices, indices, 0],
        [
            [0.0000000e00, 0.0000000e00, 1.0000000e00, 1.0000000e00],
            [2.1328260e-01, 9.8938406e-01, 6.8952787e-08, 1.8248732e-11],
            [8.7120086e-01, 5.8410158e-03, 2.0007190e-05, 4.3070755e-12],
        ],
    )


def test_aniso():
    z, y, x = np.mgrid[0:1, -25:25, -25:25]
    r = np.linalg.norm((x, y), axis=0)
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.ones_like(x)
    models = pig.Models("aliev1996simple")
    states = models.resting_states(inhom, Nframes=3)
    states[0, r < 3, 0] = 1
    diffusivity = pig.diffusivity_matrix(Df=1.0, Ds=0.2, f=[1, 2, 3])
    weights = models.weights(dz, dy, dx, inhom, diffusivity=diffusivity)
    Nt, dt = 399, 0.05
    for i in range(states.shape[0] - 1):
        states[i + 1] = models.run(inhom, weights, states[i], Nt=Nt, dt=dt)
    assert np.allclose(states[1, 0, 10, 15, 0], 0.15971428)
    assert np.allclose(states[1, 0, 15, 34, 0], 0.3017956)
    assert np.allclose(states[2, 0, 37, 33, 0], 0.1480847)
    assert np.allclose(states[2, 0, 25, 19, 0], 0.00012683906)


@pytest.mark.skipif(
    "SKIP_DOUBLE" in os.environ,
    reason="Skipping tests with double precision via SKIP_DOUBLE",
)
def test_double():
    z, y, x = np.mgrid[0:1, 0:1, 0:1]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.ones_like(x)
    models1 = pig.Models("aliev1996simple")
    models2 = pig.Models("aliev1996simple", double_precision=True)
    states: np.ndarray[Any, Any] = models2.resting_states(inhom).astype(np.float64)
    states[0, ..., 0] = 0.4
    weights = models2.weights(dz, dy, dx, inhom, diffusivity=0)
    assert states[0].ndim == 4
    states_single = models1.run(inhom, weights, states[0], Nt=199, dt=0.1)
    states_double = models2.run(inhom, weights, states[0], Nt=199, dt=0.1)
    assert states_single.ndim == 4
    assert states_double.ndim == 4
    assert states_single.dtype == np.float32
    assert states_double.dtype == np.float64
    assert np.allclose(states_single, states_double)


@pytest.mark.skipif(
    "SKIP_DOUBLE" in os.environ,
    reason="Skipping tests with double precision via SKIP_DOUBLE",
)
def test_precision_decay():
    pig.Models.available["mult"] = ModelDefinition(
        name="Multiply",
        description="",
        dois=[],
        variables={"u": 1.0},
        diffusivity={},
        parameters={"f": 0.1},
        code=r"""
            Real uf = u*f;
            *_new_u = uf > 0 ? uf : u;
        """,
    )
    inhom = np.full((1, 1, 1), 1, dtype=int)

    for d in [False, True]:
        models = pig.Models(double_precision=d)
        models.add("mult", f=0.1)  # if inhom == 1: divide by 10
        models.add("mult", f=2)  # if inhom == 2: multiply by 2

        states = models.resting_states(inhom, Nframes=3)
        weights = models.weights(1, 1, 1, inhom, diffusivity=0)

        # inhom == 1
        states[1] = models.run(inhom * 1, weights, states[0], Nt=1000, dt=1)
        small = states[1].astype(np.float64)
        if d:
            assert np.all(small < 1e-300)
        else:
            assert np.all(small > 1e-100)

        # inhom == 2
        states[2] = models.run(inhom * 2, weights, states[1], Nt=1, dt=1)
        assert states[2] > 1.1 * states[1]
        assert states[2] < 3.0 * states[1]


@pytest.mark.skipif(
    "SKIP_DOUBLE" in os.environ,
    reason="Skipping tests with double precision via SKIP_DOUBLE",
)
def test_precision_conflict():
    z, y, x = np.mgrid[0:1, 0:1, 0:1]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.ones_like(x)
    models = pig.Models("aliev1996simple", double_precision=True)
    states = models.resting_states(inhom, Nframes=2)
    weights = models.weights(dz, dy, dx, inhom, diffusivity=0)
    with pytest.warns(DeprecationWarning), pytest.raises(AssertionError):
        states = pig.run(
            models,
            inhom,
            weights,
            states[0],
            Nt=100,
            dt=0.1,
            double_precision=False,
        )


def test_dt():
    z, y, x = np.mgrid[0:1, 0:1, 0:1]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.ones_like(x)
    models = pig.Models("aliev1996simple")
    states = models.resting_states(inhom, Nframes=3)
    states[0, ..., 0] = 0.4
    weights = models.weights(dz, dy, dx, inhom, diffusivity=0)
    dt = 0.0001
    states[1] = models.run(inhom, weights, states[0], Nt=100, dt=dt)
    states[2] = models.run(inhom, weights, states[0], Nt=10, dt=dt * 10)
    assert np.allclose(states[1], states[2], rtol=1e-3, atol=1e-5)
    assert np.allclose(
        states.squeeze(),
        [
            [4.0000001e-01, 0.0000000e00],
            [4.0485540e-01, 4.8295802e-05],
            [4.0485030e-01, 4.8268699e-05],
        ],
    )



"""
Schema to define the diffusivity matrix
---------------------------------------
"""

from __future__ import annotations

from pathlib import Path
from typing import Any

import numpy as np
from pydantic import BaseModel, ConfigDict, Field

from pigreads.schema.basic import Vector3D

# pylint: disable=import-outside-toplevel


class DiffusivityFile(BaseModel):
    """
    Diffusivity matrix read from a file.

    The file should contain a 4D array with the diffusivity values. The array
    should have shape ``(Nz, Ny, Nx, 6)`` where the last dimension corresponds to
    the six independent components of the diffusivity matrix.

    :see: :py:func:`pigreads.diffusivity.diffusivity_matrix`\
          and :py:mod:`numpy.lib.format`
    """

    file: Path = Field(..., examples=[Path("path/to/diffusivity.npy")])
    "Path to the file containing the diffusivity matrix."
    model_config = ConfigDict(extra="forbid", use_attribute_docstrings=True)
    "Configuration for :py:class:`pydantic.BaseModel`."

    def __call__(self) -> np.ndarray[Any, Any]:
        """
        Read the diffusivity matrix from the file.

        :return: Diffusivity matrix.
        """
        array: np.ndarray[Any, Any] = np.load(self.file)
        assert array.ndim == 4, "Diffusivity matrix should be 4D"
        assert array.shape[-1] == 6, "Diffusivity matrix should have 6 components"
        return array


class DiffusivityParams(BaseModel):
    """
    Diffusivity matrix defined by parameters.

    The parameters are passed on to :py:func:`pigreads.diffusivity.diffusivity_matrix`.
    """

    f: Vector3D | None = Field(None, examples=[Vector3D(x=0, y=0, z=0)])
    "Main direction of diffusion."
    n: Vector3D | None = Field(None, examples=[Vector3D(x=0, y=0, z=0)])
    "Direction of weakest diffusion."
    Df: float = Field(..., examples=[1.0, 0.1])
    "Diffusivity in the direction of the fibres."
    Ds: float | None = Field(None, examples=[1.0, 0.1, None])
    "Diffusivity in the fibre sheets."
    Dn: float | None = Field(None, examples=[1.0, 0.1, None])
    "Diffusivity in the direction normal to the fibre sheets."
    model_config = ConfigDict(extra="forbid", use_attribute_docstrings=True)
    "Configuration for :py:class:`pydantic.BaseModel`."

    def __call__(self) -> np.ndarray[Any, Any]:
        """
        Create the diffusivity matrix from the parameters.

        :return: Diffusivity matrix.
        """
        from pigreads import diffusivity_matrix

        kwargs: dict[str, Any] = {"Df": self.Df, "Ds": self.Ds, "Dn": self.Dn}
        if self.f is not None:
            kwargs["f"] = self.f.tuple
        if self.n is not None:
            kwargs["n"] = self.n.tuple
        return diffusivity_matrix(**kwargs)


Diffusivity = DiffusivityFile | DiffusivityParams
"""
The diffusivity matrix is a command to define the diffusivity matrix in the
simulation.

One of:

- :py:class:`DiffusivityFile`
- :py:class:`DiffusivityParams`
"""

__all__ = [
    "Diffusivity",
    "DiffusivityFile",
    "DiffusivityParams",
]



"""
Classes to manage models
------------------------
"""

from __future__ import annotations

import re
from collections.abc import Iterator
from typing import Any, ClassVar

import numpy as np

from pigreads import core as _core
from pigreads.diffusivity import diffusivity_matrix
from pigreads.helper import ModelParameters
from pigreads.kernels import PREDEFINED_MODELS, TEMPLATE
from pigreads.schema.model import ModelDefinition, ModelEntry


class Models:
    """
    This class stores the models to be used in a Pigreads simulation. The
    models are defined by OpenCL code. The class variable :py:attr:`available`
    is a dictionary of all available models.

    Models are added to an instance of this class using the :py:meth:`add`
    method, or as a list of tuples in the constructor. The order of the models
    is the order in which they are used in the simulation. ``models[0]`` is used
    at ``inhom`` values of 1, ``models[1]`` at 2, etc.

    :param tuples_key_kwargs: A list of tuples with the model key and keyword \
    arguments encoding parameter names and values to be passed to the \
    :py:meth:`add` method. If a string is given, it is treated the name of the \
    first model to be added with no keyword arguments. If ``None``, no \
    models are added.

    :param double_precision: If ``True``, use double precision for calculations.
    :var double_precision: If ``True``, use double precision for calculations.
    :vartype double_precision: bool
    """

    available: ClassVar[dict[str, ModelDefinition]] = {
        k: ModelDefinition(**v) for k, v in PREDEFINED_MODELS.items()
    }
    """
    Dictionary of all available models.
    """

    def __init__(
        self,
        tuples_key_kwargs: list[tuple[str, dict[str, float]]] | str | None = None,
        double_precision: bool = False,
    ):
        self.double_precision: bool = double_precision
        self._core: _core.Models | None = None

        if tuples_key_kwargs is None:
            tuples_key_kwargs = []
        elif isinstance(tuples_key_kwargs, str):
            tuples_key_kwargs = [(tuples_key_kwargs, {})]
        for key, kwargs in tuples_key_kwargs:
            self.add(key, **kwargs)

    @property
    def core(self) -> _core.Models:
        """
        Get the link to the core implementation of this class.
        """
        if not isinstance(self._core, _core.Models):
            self._core = _core.Models(double_precision=self.double_precision)
        return self._core

    def __len__(self) -> int:
        """
        The number of models added to the list of models.

        .. include in docs
        """
        return len(self.core)

    def get_key(self, imodel: int) -> str:
        """
        Get the key of the model with the given index.

        :param imodel: Index of the model.
        """
        return self.core.get_key(imodel)

    def get_definition(self, imodel: int) -> ModelDefinition:
        """
        Get the definition of the model with the given index.

        :param imodel: Index of the model.
        """
        return self.available[self.get_key(imodel)]

    def get_parameter(self, imodel: int, iparam: int) -> float:
        """
        Get the parameter with the given indices.

        :param imodel: Index of the model.
        :param iparam: Index of the parameter.
        :return: Parameter value.
        """
        return self.core.get_parameter(imodel, iparam)

    def set_parameter(self, imodel: int, iparam: int, value: float) -> None:
        """
        Set the parameter with the given indices.

        :param imodel: Index of the model.
        :param iparam: Index of the parameter.
        :param value: New parameter value.
        """
        self.core.set_parameter(imodel, iparam, value)

    def get_entry(self, imodel: int) -> ModelEntry:
        """
        Get the model entry with the given index linked to the core
        implementation to read and change model parameters.

        :param imodel: Index of the model.
        :return: Model entry with a :py:class:`pigreads.helper.ModelParameters` view.
        """
        model = ModelEntry(key=self.get_key(imodel))
        model.parameters = ModelParameters(models=self, imodel=imodel)
        return model

    def __getitem__(self, imodel: int) -> ModelEntry:
        """
        Get the model entry with the given index linked to the core
        implementation to read and change model parameters.

        :param imodel: Index of the model.
        :return: Model entry with a :py:class:`pigreads.helper.ModelParameters` view.

        .. include in docs
        """
        return self.get_entry(imodel)

    def __iter__(self) -> Iterator[ModelEntry]:
        """
        Get an iterator of model entries linked to the core
        implementation to read and change model parameters.

        :return: Iterator of model entries.

        .. include in docs
        """
        for imodel in range(len(self)):
            yield self.get_entry(imodel)

    @property
    def block_size(self) -> tuple[int, int, int]:
        """
        Local work size for running OpenCL kernels.

        The ``block_size`` parameter corresponds to OpenCL's ``localWorkSize`` and
        defines the size of a small block of the domain that is processed together by
        the OpenCL platform (typically the GPU).

        If chosen too small, the performance will be suboptimal. If chosen too
        large, the OpenCL kernel will not run. The default value is ``(1, 8, 8)``,
        i.e., a block of unit width in z, and ``8x8`` in the y and x
        dimensions. This is a good compromise for most applications.

        Tweaking this value can lead to significant performance improvements.
        """
        return self.core.get_block_size()

    @block_size.setter
    def block_size(self, block_size: tuple[int, int, int]) -> None:
        """
        Set the local work size for running OpenCL kernels.

        :param block_size: Local work size.
        """
        self.core.set_block_size(block_size)

    def add(self, key: str, **parameters: Any) -> None:
        """
        Select and enable a model with given parameters.

        :param key: The key of the model to be added.
        :param parameters: Parameter names and their values to be passed to the model.
        """
        model_id = self.core.get_number_definitions()
        model_def = self.available[key]
        self.core.add(
            key,
            self.code(key, model_id),
            len(model_def.variables),
            np.array(list(model_def(**parameters).values())),
        )

    def code(self, key: str, model_id: int) -> str:
        """
        OpenCL kernel source code defining a model.

        :param key: The key of the model.
        :param model_id: An integer identifying the model.
        """
        code = TEMPLATE.render(
            model_def=self.available[key],
            key=key,
            model_id=model_id,
        )

        if self.double_precision:
            return code
        return re.sub(
            r"""
            (?<![\w.])
            (
                (?:\d+\.\d*|\.\d+|\d+\.)
                (?:[eE][+-]?\d+)?
                |
                \d+[eE][+-]?\d+
            )
            (?![fFdD\w])
        """,
            r"\1f",
            code,
            flags=re.VERBOSE,
        )

    @property
    def Nv(self) -> int:  # pylint: disable=invalid-name
        """
        Maximum number of state variables in the models.
        """
        return self.core.Nv

    @property
    def dtype(self) -> type:
        """
        Data type to use for calculations,
        i.e., single or double precision floating point numbers.
        """
        dtype: type = np.double if self.double_precision else np.single
        return dtype

    def resting_states(
        self,
        inhom: np.ndarray[Any, Any],
        Nframes: int = 1,  # pylint: disable=invalid-name
        dtype: type | None = None,
    ) -> np.ndarray[Any, Any]:
        """
        Create an array of states and fill the first frame with the resting
        values of the models depending on the ``inhom`` values.

        :param inhom: A 3D array with integer values, encoding which model to use at each point. \
                Its value is zero for points outside the medium and one or more for points inside. \
                Values larger than zero are used to select one of multiple models: \
                1 for ``models[0]``, 2 for ``models[1]``, etc.
        :param Nframes: The number of frames in time.
        :param dtype: Data type of the arrays, \
                i.e., single or double precision floating point numbers.
        :return: A 5D array of shape (Nframes, Nz, Ny, Nx, Nv).
        """

        if dtype is None:
            dtype = self.dtype

        model_count = len(self)
        assert model_count > 0, "must add at least one model"
        assert inhom.ndim == 3
        inhom = inhom.astype(int)
        mask = inhom > 0
        states: np.ndarray[Any, Any] = np.full(
            (Nframes, *inhom.shape, self.Nv), np.nan, dtype=dtype
        )
        for imodel in range(len(self)):
            model_def = self.get_definition(imodel)
            for iv, resting in enumerate(model_def.variables.values()):
                states[0, mask * ((inhom - 1) % model_count == imodel), iv] = resting
        states[:, ~mask, :] = np.nan
        return states

    def weights(
        self,
        dz: float = 1.0,
        dy: float = 1.0,
        dx: float = 1.0,
        inhom: np.ndarray[Any, Any] | None = None,
        diffusivity: np.ndarray[Any, Any] | float = 1.0,
    ) -> np.ndarray[Any, Any]:
        """
        Calculate the weights for the diffusion term in the reaction-diffusion
        equation.

        :param dz: The distance between points in the z-dimension,\
                see :py:func:`pigreads.helper.deltas`.
        :param dy: The distance between points in the y-dimension.
        :param dx: The distance between points in the x-dimension.
        :param inhom: A 3D array with integer values, encoding which model to use at each point. \
                Its value is zero for points outside the medium and one or more for points inside. \
                If ``None``, all points are considered inside the medium.
        :param diffusivity: The diffusivity matrix, \
                see :py:func:`pigreads.diffusivity.diffusivity_matrix`. \
                If a scalar is given, the matrix is isotropic with the same value in all directions.
        :return: Weight matrix for the diffusion term, A 5D array of shape (1, Nz, Ny, Nx, 19).
        """

        assert dz > 0
        assert dy > 0
        assert dx > 0

        if inhom is None:
            inhom = np.ones(shape=(1, 1, 1), dtype=self.dtype)
        assert inhom.ndim == 3

        mask = np.ascontiguousarray(inhom, dtype=np.int32) > 0
        mask.shape = (1, *mask.shape, 1)

        diffusivity = np.ascontiguousarray(diffusivity, dtype=self.dtype)
        assert isinstance(diffusivity, np.ndarray)
        if diffusivity.ndim == 1:
            diffusivity = diffusivity_matrix(
                Df=float(diffusivity.item()), dtype=self.dtype
            )
        assert diffusivity.ndim == 4
        assert diffusivity.shape[-1] == 6

        return self.core.weights(dz, dy, dx, mask, diffusivity)

    def run(
        self,
        inhom: np.ndarray[Any, Any],
        weights: np.ndarray[Any, Any],
        states: np.ndarray[Any, Any],
        stim_signal: np.ndarray[Any, Any] | None = None,
        stim_shape: np.ndarray[Any, Any] | None = None,
        Nt: int = 1,  # pylint: disable=invalid-name
        dt: float = 0.001,
    ) -> np.ndarray[Any, Any]:
        """
        Run a Pigreads simulation.

        :param inhom: A 3D array with integer values, encoding which model to use at each point. \
                Its value is zero for points outside the medium and one or more for points inside. \
                Values larger than zero are used to select one of multiple models: \
                1 for ``models[0]``, 2 for ``models[1]``, etc.
        :param weights: The weights for the diffusion term, see :py:func:`weights`.
        :param states: The initial states of the simulation, a 4D array of shape \
                (Nz, Ny, Nx, Nv), see :py:func:`Models.resting_states`.
        :param stim_signal: A 3D array with the stimulus signal at each time point \
                for all variables, with shape (Nt, Ns, Nv). If ``None``, no stimulus is applied.
        :param stim_shape: A 4D array specifying the shape of the stimulus, \
                with shape (Ns, Nz, Ny, Nx). If ``None``, no stimulus is applied
        :param Nt: The number of time steps to run the simulation for.
        :param dt: The time step size.
        :return: The final states of the simulation, a 4D array of shape (Nz, Ny, Nx, Nv).
        """
        assert Nt > 0
        assert dt > 0
        assert len(self) > 0, "must add at least one model"
        assert inhom.ndim == 3

        if stim_signal is None or getattr(stim_signal, "size", 0) == 0:
            stim_signal = np.zeros((0, 0, 0, 0, 0), dtype=self.dtype)
        else:  # np.ndarray
            assert stim_signal.ndim in [2, 3]
            stim_signal = np.reshape(
                stim_signal, (stim_signal.shape[0], -1, 1, 1, self.Nv)
            )

        assert isinstance(stim_signal, np.ndarray)
        Ns = stim_signal.shape[1]  # pylint: disable=invalid-name

        if stim_shape is None or getattr(stim_shape, "size", 0) == 0:
            stim_shape = np.zeros((0, 0, 0, 0, 0), dtype=self.dtype)
        else:  # np.ndarray
            assert stim_shape.ndim in [3, 4]
            stim_shape = np.where(inhom > 0, stim_shape, 0)
            stim_shape.shape = (Ns, *stim_shape.shape[-3:], 1)

        assert isinstance(stim_shape, np.ndarray)
        assert stim_shape.shape[0] == stim_signal.shape[1]

        states = states.astype(self.dtype).copy(order="C")
        self.core.run(
            np.ascontiguousarray(np.reshape(inhom, (*inhom.shape, 1)), dtype=np.int32),
            np.ascontiguousarray(weights, dtype=self.dtype),
            states,
            np.ascontiguousarray(stim_signal, dtype=self.dtype),
            np.ascontiguousarray(stim_shape, dtype=self.dtype),
            Nt,
            dt,
        )
        return states


__all__ = [
    "Models",
]



"""
Schema for some basic data types
--------------------------------
"""

from __future__ import annotations

from pydantic import BaseModel, ConfigDict, Field


class Vector3D(BaseModel):
    """
    Vector in three dimensions in space.
    """

    x: float = Field(0, examples=[0.1, 1.0, -0.1])
    "Value in x direction."
    y: float = Field(0, examples=[0.1, 1.0, -0.1])
    "Value in y direction."
    z: float = Field(0, examples=[0.1, 1.0, -0.1])
    "Value in z direction."
    model_config = ConfigDict(extra="forbid", use_attribute_docstrings=True)
    "Configuration for :py:class:`pydantic.BaseModel`."

    @property
    def tuple(self) -> tuple[float, float, float]:
        """
        Representation as a tuple (x, y, z).
        """
        return self.x, self.y, self.z


class Slice(BaseModel):
    """
    Slice of an array.

    :see: :py:class:`slice`
    """

    axis: int = Field(-1, examples=[-1, 0, 3])
    "Axis to slice in."
    start: int | None = Field(None, examples=[4, -1, None])
    "Start index of the slice."
    end: int | None = Field(None, examples=[-2, 4, -1, None])
    "End index of the slice."
    step: int | None = Field(None, examples=[1, 2, -1, None])
    "Index step of the slice."
    model_config = ConfigDict(extra="forbid", use_attribute_docstrings=True)
    "Configuration for :py:class:`pydantic.BaseModel`."

    def __call__(self) -> slice:
        """
        Return a slice object.

        :return: Slice object.
        :see: :py:class:`slice`
        """
        return slice(self.start, self.end, self.step)


__all__ = [
    "Slice",
    "Vector3D",
]
